local DFUtil = require("DFCommon.Util")

local m = {
    publicKey = nil,    
    productIds = {},
    receiptVerifyURL = "https://services.moaicloud.com/dfmobile/shopverify",
    productCatalog = {},
    purchasesAllowed = false,
    purchaseCallback = nil,
    failureCallback = nil,
    refundCallback = nil,    
    forceFakePurchases = false,
    tUnconsumedProducts = {}, -- id and purchaseinfo pair.  This is needed to consume purchases for Google Play store

    -- Error codes for failureCallback
    kPURCHASE_DECLINED = 0,
    kPURCHASE_CANCELLED = 1,
}

local kBILLINGPROVIDER_UNDEFINED = -1
-- KSC: need to change this depending on google play build or amazon build
-- kind of sucks, but there's no real good way for moai to recognize which one to use
MOAIEnvironment.androidBillingProvider = MOAIBilling.BILLING_PROVIDER_GOOGLE
--MOAIEnvironment.androidBillingProvider = MOAIBilling.BILLING_PROVIDER_AMAZON

-- init
-- Initialzies the lib, begins polling the system for IAP related data.
--
-- publicKey: The public key string for this app. Found in the app store mgmt web page. Currently Android only.
-- productIds: A list of product id strings that match the list in the game's app store.
-- purchaseCallback: The function to call when a purchase is successfully completed or restored.
-- failureCallback: The function called when a purchase fails. An error code will tell why the purchase failed.
-- refundCallback: The function to call when a purchase is refunded. Currently Android only.
function m.init(publicKey, productIds, receiptVerifyURL, purchaseCallback, failureCallback, refundCallback)
    m.publicKey = publicKey    
    m.productIds = productIds
    m.receiptVerifyURL = receiptVerifyURL
    m.purchaseCallback = purchaseCallback
    m.failureCallback = failureCallback
    m.refundCallback = refundCallback
    
    if MOAIEnvironment.osBrand == "iOS" then
        MOAIBilling.setListener ( MOAIBilling.PAYMENT_QUEUE_TRANSACTION, m._onPaymentQueueTransaction_iOS )
        MOAIBilling.setListener ( MOAIBilling.PRODUCT_REQUEST_RESPONSE, m._productRequestResponseCallback_iOS )

    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE then
        MOAIBilling.setListener( MOAIBilling.CHECK_BILLING_SUPPORTED, m._purchasePermissionsCallback )
        MOAIBilling.setListener( MOAIBilling.PURCHASE_RESPONSE_RECEIVED, m._onPurchaseResponseReceived_Google )
                
        MOAIBilling.setBillingProvider( MOAIBilling.BILLING_PROVIDER_GOOGLE )
        MOAIBilling.setPublicKey(publicKey)
    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_AMAZON then
        MOAIBilling.setBillingProvider( MOAIBilling.BILLING_PROVIDER_AMAZON )

        MOAIBilling.setListener( MOAIBilling.PURCHASE_RESPONSE_RECEIVED, m._onPurchaseResponseReceived_Amazon )
        MOAIBilling.setListener( MOAIBilling.PURCHASE_STATE_CHANGED, m._onPurchaseStateChanged_Amazon )        
        MOAIBilling.setListener( MOAIBilling.ITEM_RESPONSE_RECEIVED, m._onItemResponseReceived_Amazon )        

    else
        -- fake it till you make it!
    end
           
    m._updateStore()    
end

-- arePurchasesAllowed
-- Returns true if purchases are allowed by the system, false otherwise.
-- Clients should call this function before presenting any IAP UI to the player.
--
-- On some systems it takes time to determine this so may be false right away
-- an true later on. Similarly, it can change state at any moment in the app lifecycle.
function m.arePurchasesAllowed()
    return m.purchasesAllowed
end

-- getProductCatalog
-- Returns the product catalog for all items in the store.
-- Currently, only fully fleshed out on iOS.
--
-- On some systems it takes time to determine this so may be false right away
-- an true later on. Similarly, it can change state at any moment in the app lifecycle.
function m.getProductCatalog()
    return m.productCatalog
end

-- purchaseItem
-- Request that the system initiate a purchase of an item in the store.
-- Must only be called based on direct user input.
-- If successful, calling this function will result in the purchaseCallback
-- being triggered, but it may be some time in the future.
--
-- productId: Identifier of product to be purchased.
function m.purchaseItem(productId)  
    if m.productCatalog[productId] ~= nil then
        if MOAIEnvironment.osBrand == "iOS" and not m.forceFakePurchases then
            MOAIBilling.requestPaymentForProduct(productId)
        elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE and not m.forceFakePurchases then
            local result = MOAIBilling.purchaseProduct(productId)
            if result ~= MOAIBilling.BILLINGV3_RESPONSE_RESULT_OK then
                m.failureCallback(productId, m.kPURCHASE_DECLINED)	
                return false
            end
        elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_AMAZON and not m.forceFakePurchases then
            local result = MOAIBilling.requestPurchase(productId)
            -- we don't invoke the failure callback here.  it's on a callback
            if not result then
                return false
            end
        else
            m._fakePurchase(productId)
        end        
    end
    return true
end

-- restorePurchases
-- Request that the relevant app backend resend any previous non-consumable purchases.
-- Restored purchases will appear as regular purchaseCallbacks.
--
function m.restorePurchases()
    if MOAIEnvironment.osBrand == "iOS" then
        MOAIBilling.restoreCompletedTransactions()
    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE then
        local sResultJSON = MOAIBilling.getPurchasedProducts()
        if sResultJSON then
            local tPurchaseInfo = MOAIJsonParser.decode(sResultJSON)
            if tPurchaseInfo then
                if tPurchaseInfo.purchaseData and type(tPurchaseInfo.purchaseData) == 'table' then                    
                    for i, tInfo in ipairs(tPurchaseInfo.purchaseData) do
                        local tRetTable = {}
                        tRetTable[1] = tInfo -- this is the table state that we expect
                        local sProductJSON = MOAIJsonParser.encode(tRetTable)
                        m._onPurchaseResponseReceived_Google(MOAIBilling.BILLING_RESULT_SUCCESS, sProductJSON)
                    end
                end
            end
        end
    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_AMAZON then
        MOAIBilling.restoreTransactions() -- this will just call the state changed callbacks
    end
end

-- consumePurchases
-- For consume consumable products so that they can be purchased again
-- This is a Google Play store requirement
-- Be VERY careful to only pass consumable items to this as unconsumable items shouldn't be available to purchase again
function m.consumePurchase(id)
    if MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE then
        -- find the product
        if id then
            for i, rInfo in ipairs(m.tUnconsumedProducts) do
                if rInfo.productId == id then
                    if rInfo.tResult and rInfo.tResult.purchaseToken then
                        MOAIBilling.consumePurchaseSync(rInfo.tResult.purchaseToken) -- CONSUME!!!
                        table.remove(m.tUnconsumedProducts, i)
                        break
                    end
                end
            end
        end
    end
end

-- shutdown
-- Terminates the IAP system
function m.shutdown()
    m.storeUpdateThread = nil
    m.fakePurchaseThread = nil
end

-- Private Functions

function m._updateStore()
    m.storeUpdateThread = MOAICoroutine.new()
	m.storeUpdateThread:run(
        function()
            while true do                
                if MOAIEnvironment.osBrand == "iOS" then
                    -- some platforms have a delayed response, so mimic that on iOS
                    DFUtil.sleep(0.5)
                    m._purchasePermissionsCallback( MOAIBilling.canMakePayments() )
                elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE then
                    -- fire off a request to see if we can make purchses on this device
                    m.purchasesAllowed = MOAIBilling.checkInAppSupported()
                    -- wait a few seconds to see if our billing status has changed
                    -- TODO: early out when we get the callback
                    DFUtil.sleep(2)
                elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_AMAZON then
                    m.purchasesAllowed = MOAIBilling.checkBillingSupported()
                    -- wait a few seconds to see if our billing status has changed
                    -- TODO: early out when we get the callback
                    DFUtil.sleep(2)                    
                else                    
                    -- some platforms have a delayed response, so mimic that in the faker path.
                    DFUtil.sleep(0.5)
                    -- test platforms can always make IAP                    
                    m._purchasePermissionsCallback( true )
                end
                
                if m.purchasesAllowed then
                    if MOAIEnvironment.osBrand == "iOS" then
                        MOAIBilling.requestProductIdentifiers( m.productIds )
                    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_GOOGLE then
                        -- not sure what to do here, pending info from Moai
                        -- for now, just fill out a basic catalog.
                        local sProductJSON = MOAIBilling.requestProductsSync(m.productIds, MOAIBilling.BILLINGV3_PRODUCT_INAPP)
                        if sProductJSON then
                            local tReturnProductInfo = MOAIJsonParser.decode(sProductJSON)
                            if tReturnProductInfo then
                                m.productCatalog = {}    
                                for i, sProductJSON in ipairs(tReturnProductInfo) do
                                    local productInfo = MOAIJsonParser.decode(sProductJSON) 
                                    if productInfo and productInfo.productId then
                                        m.productCatalog[productInfo.productId] = {
                                            id = productInfo.productId,
                                            title = productInfo.title,
                                            description = productInfo.description,
                                            price = productInfo.price,
                                            localizedPrice = productInfo.price,
                                            priceLocale = "", -- KSC: We don't get this from Google
                                        }
                                    end
                                end
                            end
                        else
                            print("NO PRODUCTS!!!")
                        end
                    elseif MOAIEnvironment.osBrand == "Android" and MOAIEnvironment.androidBillingProvider == MOAIBilling.BILLING_PROVIDER_AMAZON then
                        -- the request for products get sent asynchonously
                        MOAIBilling.requestSkusData(m.productIds)
                    else
                        DFUtil.sleep(0.5)
                        m._createFakeCatalog()
                    end
                end
                
                -- recheck the billing status every so often as users can change it outside the app
                DFUtil.sleep(300)
            end
        end
    )
end

-- Universal Callbacks

-- _purchasePermissionsCallback
-- Called when device purchase permission state changes.
-- On many systems, this callback is async.
--
-- allowed: State of purchase system at time of callback.
function m._purchasePermissionsCallback( allowed )
    m.purchasesAllowed = allowed
end

-- iOS Callbacks

-- _productRequestResponseCallback_iOS
-- Called when the store delivers a new list of items.
-- 
-- products: Table of purchaseable items and properties.
function m._productRequestResponseCallback_iOS( products )
    m.productCatalog = {}    
	for i, v in ipairs( products ) do
        m.productCatalog[v.productIdentifier] = {
            id = v.productIdentifier,
            title = v.localizedTitle,
            description = v.localizedDescription,
            price = v.price,
            localizedPrice = v.localizedPrice,
            priceLocale = v.priceLocale,            
        }
	end
end

-- _onPaymentQueueTransaction_iOS
-- Called when a transaction is completed (successfully or unsuccessfully)
-- Generally, we only care about successful purchases.
--
-- transaction: Table describing all relevant state for the transaction.
function m._onPaymentQueueTransaction_iOS( transaction )
    if (transaction.transactionState == MOAIBilling.TRANSACTION_STATE_PURCHASED or transaction.transactionState == MOAIBilling.TRANSACTION_STATE_RESTORED) and m.purchaseCallback then        
        if m.receiptVerifyURL then
            local recieptThread = MOAICoroutine.new()
            recieptThread:run(
                function()
                    local localReceiptTask = MOAIHttpTask.new() 
                    localReceiptTask:setTimeout( 5 )      
                    
                    local reqBuffer = MOAIDataBuffer.new()
                    reqBuffer:base64Encode(transaction.transactionReceipt)                    
                    
                    localReceiptTask:httpPost(m.receiptVerifyURL, reqBuffer:getString(), nil, nil, true)
                    local receiptResults = MOAIJsonParser.decode(localReceiptTask:getString())
                    local connectedToServer = localReceiptTask:getResponseCode() == 400 or localReceiptTask:getResponseCode() == 200
                    if connectedToServer and (receiptResults == nil or receiptResults.status ~= 0) then
                        if m.failureCallback then
                            m.failureCallback(transaction.payment.productIdentifier, m.kPURCHASE_DECLINED)
                        end
                    else
                        if m.purchaseCallback then
                            m.purchaseCallback(transaction.payment.productIdentifier, transaction.payment.quantity, transaction.transactionState == MOAIBilling.TRANSACTION_STATE_RESTORED)
                        end
                    end
                end
            )
        else
            m.purchaseCallback(transaction.payment.productIdentifier, transaction.payment.quantity, transaction.transactionState == MOAIBilling.TRANSACTION_STATE_RESTORED)
        end

        
    elseif transaction.transactionState == MOAIBilling.TRANSACTION_STATE_FAILED and m.failureCallback then
        m.failureCallback(transaction.payment.productIdentifier, m.kPURCHASE_DECLINED )
    elseif transaction.transactionState == MOAIBilling.TRANSACTION_STATE_CANCELLED and m.failureCallback then
        m.failureCallback(transaction.payment.productIdentifier, m.kPURCHASE_CANCELLED )
    end
end

-- _onPurchaseResponseReceived_Android
-- Called when a purchase processing request is made.
-- This is not the actual pruchase, but a status that purchase action was taken.
-- 
-- code: The status of the request response.
-- id: The id of the product in question.
function m._onPurchaseResponseReceived_Android( code, id )
 	if code == MOAIBilling.BILLING_RESULT_SUCCESS then
		print( "purchase request received" )
	elseif code == MOAIBilling.BILLING_RESULT_USER_CANCELED then
		print( "user canceled purchase" )
	else
		print( "purchase failed" )
	end
end

-- _onPurchaseResponseReceived_Google
-- Called when a purchase processing request is made.
-- This is not the actual pruchase, but a status that purchase action was taken.
-- 
-- code: The status of the request response.
-- id: The id of the product in question.
function m._onPurchaseResponseReceived_Google( code, resultJSONString )
    if MOAIEnvironment.androidBillingProvider ~= MOAIBilling.BILLING_PROVIDER_GOOGLE then
        return
    end
 	if code == MOAIBilling.BILLING_RESULT_SUCCESS then
		print( "purchase request received" )
	elseif code == MOAIBilling.BILLING_RESULT_USER_CANCELED then
		print( "user canceled purchase" )
	else
		print( "purchase failed" )
	end

    
    local tResult = nil
    local id = ''    

 	if code == MOAIBilling.BILLING_RESULT_SUCCESS then
        if resultJSONString then
            tResult = MOAIJsonParser.decode(resultJSONString)        
            if tResult and tResult[1] then               
                if type(tResult[1]) == 'table' then
                    tResult = tResult[1]
                elseif type(tResult[1] == 'string') then
                    tResult = MOAIJsonParser.decode(tResult[1]) --ug getting the string from google, the second part doesn't get decoded
                end
            end
        end
    
        if tResult then
            if tResult.productId then
                id = tResult.productId
            end

            if tResult.purchaseState then
                if tResult.purchaseState == MOAIBilling.BILLING_PURCHASE_STATE_ITEM_PURCHASED then
                    if m.purchaseCallback then
                        -- we save out the purchase info.  For google billing, consumable items/managed products must be "consumed".  The callback is responsible for consuming the item
                        local tInfo = {
                            productId = id,
                            tResult = tResult, -- this will contain all info about the purchase
                        }
                        table.insert(m.tUnconsumedProducts, tInfo)   
                        m.purchaseCallback(id, 1, false)  
                    end
                elseif tResult.purchaseState == MOAIBilling.BILLING_PURCHASE_STATE_ITEM_REFUNDED then
                    if m.refundCallback then
                        m.refundCallback(id, 1, true) 
                    end
                elseif tResult.purchaseState == MOAIBilling.BILLING_RESULT_USER_CANCELED then
                    m.failureCallback(id, m.kPURCHASE_CANCELLED)
                end
            end
        end
    end
    m.failureCallback(id, m.kPURCHASE_CANCELLED) -- defaultinf to cancelled for a lack of a better default
end


-- _onPurchaseResponseReceived_Amazon
-- Called when a purchase processing request is made.
-- This is not the actual pruchase, but a status that purchase action was taken.
-- 
-- code: The status of the request response.
-- id: The id of the product in question.
function m._onPurchaseResponseReceived_Amazon(code, id)
    if MOAIEnvironment.androidBillingProvider ~= MOAIBilling.BILLING_PROVIDER_AMAZON then
        return
    end
 	if code ~= MOAIBilling.BILLING_RESULT_SUCCESS then    
        m.failureCallback(id, m.kPURCHASE_DECLINED)	
    end
    -- success is only handled by the product state changed callback
end

-- _onPurchaseStateChanged_Amazon
-- Called when the state of a item has changed, e.g. purchased a product
-- code: the status of the purchase
-- productID: the productID of the sku purchased
-- orderReceipt: data to be checked if we are verifying
function m._onPurchaseStateChanged_Amazon(code, productID, orderReceipt)
    if productID then
        if code == MOAIBilling.BILLING_PURCHASE_STATE_ITEM_PURCHASED then
            m.purchaseCallback(productID, 1, false)
        elseif code == MOAIBilling.BILLING_PURCHASE_STATE_ITEM_REFUNDED then
            if m.refundCallback then
                m.refundCallback(productID, 1, true) 
            end            
        end
    end
end

-- _onItemResponseReceived_Amazon
-- Called when we get information about skus
-- This is where we populate the catalog
-- 
-- itemInfoJSON: The json of the sku info
function m._onItemResponseReceived_Amazon(sProductJSON)
    if sProductJSON and type(sProductJSON) == 'string' then
        local tReturnProductInfo = MOAIJsonParser.decode(sProductJSON)
        if tReturnProductInfo then
            if type(tReturnProductInfo) == 'table' then
                m.productCatalog = {}    
                for i, productInfo in ipairs(tReturnProductInfo) do
                    if productInfo.sku then
                        m.productCatalog[productInfo.sku] = {
                            id = productInfo.sku,
                            title = productInfo.title,
                            description = productInfo.description,
                            price = productInfo.price,
                            localizedPrice = productInfo.price,
                            priceLocale = "", -- KSC: We don't get this from Google
                        }
                    end
                end
            end
        end
    end
end

-- Faker Functions (for platforms w/ out stores)

-- _createFakeCatalog
-- Takes the list of passed in product IDs and creates a fake catalog from them.
-- No data created this way should be trusted except for the structure of said data.
function m._createFakeCatalog()    
    m.productCatalog = {}
    -- create fake version of full data
    for i,v in ipairs(m.productIds) do
        m.productCatalog[v] = {
            id = v,
            title = "Test Product",
            description = "This is for testing!",
            price = 0.99,
            localizedPrice = "$0.99",
            priceLocale = "FKM",
        }
    end
end

-- _fakePurchase
-- Pretends to make a purchase w/ the same latency pattern as a real store.
--
-- productId: The product id to be purchased.
function m._fakePurchase(productId)
    if m.fakePurchaseThread == nil then
        m.fakePurchaseThread = MOAICoroutine.new()
        m.fakePurchaseThread:run(
            function()
                DFUtil.sleep(1)
                m.purchaseCallback(productId, 1, false)            
                m.fakePurchaseThread = nil
            end
        )
    end
end

return m