local DFUtil = require "DFCommon.Util"
local url = require "moai.url"

local kDATA_FILE_NAME = "news.txt"

local m = {
    data = nil,             -- The active net data for use by external parties
    
    url = nil,              -- The url from which to fetch the data.
    
    pendingData = nil,      -- The latest data from the network. Call updateData() to push into data()
    httpTask = nil,         -- Non-nil if an HTTP request is currently in flight.
    pollInterval = 1800,    -- Seconds between requests for new data from the server. Should on order of minutes.
    initialDelay = 1,       -- Initial delay before the first request. Some plats take time after boot to determine network connectivity.
    doShutdown = true,      -- An shutdown request is pending. Is for internal use only.
    bHasNewData = false,
}

-- Public functions

-- init
-- Initialzies the lib, begins polling the server for new updates.
-- Data will be cached to disk to ensure they exist even if future
-- play sessions have no network connectivity.alignSprite
--
-- dataURL: The URL to the web service for this app.
-- pollInterval: Seconds between requests for new data from the cloud server.
-- initialDelay: Seconds before the first request to the cloud server.
--               Allows for platforms that need time after boot to determine network connectivity.
function m.init(dataURL, pollInterval, initialDelay)    
    m.doShutdown = false
    
    -- copy in external data
    m.url = dataURL
    
    -- override optional args
    if pollInterval ~= nil then
        m.pollInterval = pollInterval
    end
       
    if initialDelay then
        m.initialDelay = initialDelay
    end
    
    -- immediately load the last cached data if possible
    local dataBuffer = MOAIDataBuffer.new()
    -- synchronous load for client simplicity
    -- can make async if we see hitches here,
    -- but should be behind a loading screen anyway.
    dataBuffer:load(m.getDataFile())
    dataBuffer:inflate()
    -- cached data goes immediately into the valid data table
    m.data = MOAIJsonParser.decode( dataBuffer:getString() )  
    
    -- spin off a coroutine to talk to the server.
    m.pollThread = MOAICoroutine.new()
	m.pollThread:run(
        function()
            -- some platforms ping a server to establish connectivity
            -- so let's wait for a short amount of time to make sure
            -- Moai's internal network state is accurate. Otherwise,
            -- we'd potentially wait many minutes between polls.
            local initTimer = MOAITimer.new()
            initTimer:setSpan( m.initialDelay )
            initTimer:start()
            MOAICoroutine.blockOnAction(initTimer) 
        
            -- spin looking for new tasks
            while true do
                -- non mobile platforms will return CONNECTION_TYPE_WIFI even if using
                -- an Ethernet connection. Hacky but effective.
                local connectionType = MOAIEnvironment.connectionType	                	
		        if connectionType == MOAIEnvironment.CONNECTION_TYPE_WIFI or connectionType == MOAIEnvironment.CONNECTION_TYPE_WWAN then		        					
                    -- build up the full URL
                    local fullUrl = m.url .. "?"
                                        
                    if MOAIEnvironment.appVersion then
                        fullUrl = fullUrl .. "&buildId=" .. MOAIEnvironment.appVersion
                    end
                    
                    -- create and dispatch the actual task
                    m.httpTask = MOAIHttpTask.new()
                    m.httpTask:setCallback( m.httpFinish )  
                    m.httpTask:setTimeout( 5 )   
                    m.bHasNewData = false
                    m.httpTask:httpGet( fullUrl )
                end
            
                -- wait for a while or shutdown
                local timer = MOAITimer.new()
                timer:setSpan( m.pollInterval )
                timer:start()
                MOAICoroutine.blockOnAction(timer)                
            end
        end
    )
end

-- updateData
-- Copies the pendingData table into the data table.
-- Should be called by clients at a point in game where
-- it's safe to update the data.
function m.updateData()
    if m.pendingData then
        m.bHasNewData = true
        -- copy the table to keep it independent of the pendingData ref.
        m.data = DFUtil.deepCopy(m.pendingData)
        m.pendingData = nil
    end
end

function m.hasNewData()
    return m.bHasNewData
end

function m.clearNewData()
    m.bHasNewData = false
end

-- shutdown
-- Terminates the system including any pending HTTP tasks.
function m.shutdown()    
    m.pollThread = nil
    m.httpTask = nil
    m.data = nil
    m.pendingData = nil
    m.doShutdown = true
end

-- Private utility functions

-- getDataFile
--
-- returns: string path to the full data file.
function m.getDataFile()
    local docDir = MOAIEnvironment.documentDirectory
    return docDir .. "/" .. kDATA_FILE_NAME
end

-- httpFinish
-- task: the MoaiHTTPTask that generated this callback
--
-- responseCode: the HTTP response code from the task
function m.httpFinish( task, responseCode )
    if not m.doShutdown and responseCode == 200 then
        -- data come back as JSON data        
        local jsonData = task:getString()
        -- invalid request will have nil string
        if jsonData then
            -- store new data as lua table            
            m.pendingData = MOAIJsonParser.decode( jsonData )
            -- cache jsonData to HD
            local dataBuffer = MOAIDataBuffer.new()
            dataBuffer:setString( jsonData )
            -- zip it for obscurity and smaller size
            dataBuffer:deflate()
            -- save could be async, but sync for now for simplicity.
            dataBuffer:save( m.getDataFile() )            
        end        
    end
    -- clear out the task pointer
    m.httpTask = nil
end

return m