module(..., package.seeall)

local Renderer = require('Renderer')

local bDebugDontDrawMap = true

function Map:new(sMapFilename)
	
	-- PRE CONSTRUCTOR
	print("Loading map: " .. sMapFilename)
	local tMap = { }
	
	-- PROCTECTED FUNCTIONS --	
	function tMap:_init()
	
		-- Cache renderer values
		local rLayer = Renderer.getRenderLayer('Gameplay')
		
		tMap.prop = MOAIProp2D.new()
		rLayer:insertProp( tMap.prop )
		
		-- Create static layers
		tMap.layerProps = { }
		local baseX = 0
		local baseY = 0
		local numLayers = #tMap.MapData.layers
		for i=1,numLayers do
			local layer = tMap.MapData.layers[i]
			
			-- Calculate placement
			local x1 = layer.width / 2
			local x0 = -x1
			local y1 = layer.height / 2
			local y0 = -y1
			
			if i == 1 then
				tMap.baseX = x0
				tMap.baseY = y0
				
				tMap.width = layer.width
				tMap.height = layer.height
			end
			
			--print("Layer " .. layer.image .. ": " .. tostring(x0) .. ", " .. tostring(y0) .. " -> " .. tostring(x1) .. ", " .. tostring(y1))
			
			-- Create tile deck
			layer.deck = MOAIGfxQuad2D.new()
			layer.deck:setTexture( layer.image )
			layer.deck:setRect( x0, y0, x1, y1 )
			
			-- Create prop
			layer.prop = MOAIProp2D.new()
			layer.prop:setDeck( layer.deck )
			layer.prop:setPriority( layer.priority )
			layer.prop:setParent(tMap.prop)
			
			if i ~= 1 then
				local xOff = layer.width / 2 - tMap.width / 2 + layer.x
				local yOff = tMap.height / 2 - layer.height / 2 - layer.y
				layer.prop:setLoc(xOff, yOff)
			end

			if not bDebugDontDrawMap then
                rLayer:insertProp( layer.prop )
            end
			
			table.insert(tMap.layerProps, layer.prop)
		end
		
		-- Create the layout
		local layout = tMap.MapData.layout	
		
		local tileWidth = layout.tileWidth
		layout.tileHalfWidth = tileWidth / 2.0
		local tileHeight = layout.tileHeight
		layout.tileHalfHeight = tileHeight / 2.0
		
		tMap.prop:setScl(layout.scaleX, layout.scaleY)
		
		-- Load all the tile images (if requested)
		local numVisibleTileTypes = 0
		local numTileTypes = #layout.tileTypes
		for i=1,numTileTypes do
			local tileType = layout.tileTypes[i]
			if not tileType.opacity then
				tileType.opacity = layout.tileOpacity
			end
			if tileType.opacity > 0 and tileType.image then
				tileType.tileDeck = MOAIGfxQuad2D.new()
				tileType.tileDeck:setTexture( tileType.image )
				tileType.tileDeck:setRect( -layout.tileHalfWidth, -layout.tileHalfHeight, layout.tileHalfWidth, layout.tileHalfHeight )
				
				numVisibleTileTypes = numVisibleTileTypes + 1
				
				if not tMap.debugTile then
					tMap.debugTile = MOAIProp2D.new()
					tMap.debugTile:setDeck( tileType.tileDeck )
					tMap.debugTile:setLoc( 0, 0 )
					tMap.debugTile:setColor( 1, 0, 0, 0.75 )
					tMap.debugTile:setVisible(false)
					local rDebugLayer = Renderer.getRenderLayer('GameplayOverlay')
					rDebugLayer:insertProp( tMap.debugTile )
				end
			end
		end
		
		-- Calculate offsets
		tMap.baseOffsetX = tMap.baseX + layout.offsetX
		tMap.baseOffsetY = -tMap.baseY - layout.offsetY

		if numVisibleTileTypes > 0 then
			
			-- Create the debug geometry
			local index = 1
			for y=1,layout.numTilesY do
				for x=1,layout.numTilesX do
				
					local tileIndex = layout.map[index]
					local tileType = layout.tileTypes[tileIndex]
					
					if tileType.tileDeck then
					
						local layerIndex = layout.layermap[index]
						local layer = tMap.MapData.layers[layerIndex]
					
						local xOffset = -y * layout.tileHalfWidth + tMap.baseOffsetX
						local yOffset = -y * layout.tileHalfHeight + tMap.baseOffsetY
					
						local xPos = x * layout.tileHalfWidth + xOffset
						local yPos = -x * layout.tileHalfHeight + yOffset
						
						--print("Tile " .. tostring(x) .. "," .. tostring(y) .. " pos:" .. tostring(xPos) .. "," .. tostring(yPos))
					
						-- Create prop
						local tile = MOAIProp2D.new()
						tile:setDeck( tileType.tileDeck )
						tile:setLoc( xPos, yPos )
						tile:setPriority(layer.priority + 1)
						
						-- Setup tint and transparency
						local color = { r = 1, g = 1, b = 1, a = 1 }
						if tileType.tintR then color.r = tileType.tintR end
						if tileType.tintG then color.g = tileType.tintG end
						if tileType.tintB then color.b = tileType.tintB end
						if tileType.opacity < 1 then
							color.a = tileType.opacity
							tile:setBlendMode( MOAIProp2D.GL_SRC_ALPHA, MOAIProp2D.GL_ONE_MINUS_SRC_ALPHA )
						end
						
						-- Visualize the door
						if x == tileType.exitX and y == tileType.exitY then
							color.r = color.r * 0.5
							color.g = color.g * 0.5
							color.b = color.b * 0.5
						end
						
						tile:setColor( color.r, color.g, color.b, color.a )
						
						-- Add to renderer
						local rTileLayer = Renderer.getRenderLayer( tileType.renderLayer )
						if not rTileLayer then
							rTileLayer = rLayer
						end
						rTileLayer:insertProp ( tile )

					end
				
					index = index + 1
				end
			end
		end
	end
	
	-- CONSTRUCTOR
	tMap.Filename = sMapFilename
	tMap.MapData = require(sMapFilename)
	tMap._init()
	
	-- FUNCTIONS
	function tMap:getWorldCoords(x, y)
	
		local layout = tMap.MapData.layout	
		
		local xOffset = -y * layout.tileHalfWidth + tMap.baseOffsetX
		local yOffset = -y * layout.tileHalfHeight + tMap.baseOffsetY
	
		local xPos = x * layout.tileHalfWidth + xOffset
		local yPos = -x * layout.tileHalfHeight + yOffset
		
		return { x = xPos, y = yPos }
	end
	
	function tMap:getTileCoords(x, y)
		
		local layout = tMap.MapData.layout	
		local a = x * layout.tileHalfHeight + tMap.baseOffsetY * layout.tileHalfWidth - y * layout.tileHalfWidth - tMap.baseOffsetX * layout.tileHalfHeight
		local b = 2 * layout.tileHalfWidth * layout.tileHalfHeight
		local xPos = a / b
		
		local yPos = (tMap.baseOffsetY - xPos * layout.tileHalfHeight - y) / layout.tileHalfHeight
		
		local xi = math.floor(xPos + 0.5)
		local yi = math.floor(yPos + 0.5)
		
		--[[
		-- Find out whether or not the coordinate is inside the tile
		local dx = x - xi
		local dy = y - yi		
		local adx = math.abs(dx)
		local ady = math.abs(dy)
		
		a = adx / layout.tileHalfWidth
		a = math.max(math.min(a, 1), 0)
		b = (1 - a) * layout.tileHalfHeight
		]]--
		
		local c = tMap:getWorldCoords(xi, yi)
--        print("Row: "..xi.." Col: "..yi)
		return { x = xi, y = yi, cx = c.x, cy = c.y, x0 = c.x - layout.tileHalfWidth, y0 = c.y - layout.tileHalfHeight, x1 = c.x + layout.tileHalfWidth, y1 = c.y + layout.tileHalfHeight }
	end
	
	function tMap:getTilePosAtWorldPos(x, y)
		local coords = tMap:getTileCoords(x, y)
		return coords.cx, coords.cy
	end
	
	function tMap:getTileType(x, y)
		local coords = tMap:getTileCoords(x, y)
		local layout = tMap.MapData.layout	
		local index = (coords.y - 1) * layout.numTilesX + coords.x
		local tileIndex = layout.map[index]
		local tileType = layout.tileTypes[tileIndex]
		return tileType
	end
	
	function tMap:getTileTypeByName(sName)
		local layout = tMap.MapData.layout	
		local numTileTypes = #layout.tileTypes
		for i=1,numTileTypes do
			if layout.tileTypes[i].name == sName then
				return layout.tileTypes[i]
			end
		end
		return nil
	end
	
	function tMap:getTileTypeName(x, y)
		local tileType = tMap:getTileType(x, y)
		if tileType then
			return tileType.name
		else
			print("WARNING: Could not return a tile type for this x, y")
			return nil
		end
	end
	
	function tMap:getRenderPriority(x, y)
		local coords = tMap:getTileCoords(x, y)
		
		--[[
		-- DEBUG HACK
		print(tostring(x) .. ', ' .. tostring(y) .. ' -> ' .. tostring(coords.x) .. ', ' .. tostring(coords.y))
		if tMap.debugTile then
			tMap.debugTile:setLoc(coords.cx, coords.cy)
			tMap.debugTile:setVisible(true)
		end
		]]--
		
		local layout = tMap.MapData.layout	
		if coords.x < 1 or coords.x > layout.numTilesX or coords.y < 1 or coords.y > layout.numTilesY then
			print("Coordinate isn't inside the map: " .. tostring(x) .. ', ' .. tostring(y))
			return 1000
		end
		
		local index = (coords.y - 1) * layout.numTilesX + coords.x
		local layerIndex = layout.layermap[index]
		local layer = tMap.MapData.layers[layerIndex]
		local tilePriority = layer.priority
		
		local priorityOffset = y - coords.y0	
		local renderPriority = tilePriority + priorityOffset
		
		return renderPriority
	end
	
	function tMap:getRoomLevel(sName)
		local tileType = tMap:getTileTypeByName(sName)
		local nLevel = 0
		if tileType then
			nLevel = 1
			if tileType.level then
				nLevel = tileType.level
			else
				tileType.level = nLevel
			end
		end
		return nLevel
	end
	
	function tMap:setRoomLevel(sName, nLevel)
		local tileType = tMap:getTileTypeByName(sName)
		if tileType then
			tileType.level = nLevel
			
			if tileType.level == 2 then
				print("Applying layer modifiers: " .. sName)
				tMap:applyLayerMods(tileType.layerMods)
			end
		end
	end
	
	function tMap:applyLayerMods(layerMods)
		if not layerMods then
			return
		end
		
		local rLayer = Renderer.getRenderLayer('Gameplay')
		
		local numLayerMods = #layerMods
		for i=1,numLayerMods do
			local layerModIndex = layerMods[i]
			local layerMod = tMap.MapData.layerMods[layerModIndex]
			
			-- Calc priority
			local layer = tMap.MapData.layers[layerMod.layer]
			layerMod.priority = layer.priority + 1
			
			if layerMod.modType == "replace" then 
				layerMod.priority = layer.priority 
			end
			
			-- Calculate placement
			local x1 = layerMod.width / 2
			local x0 = -x1
			local y1 = layerMod.height / 2
			local y0 = -y1
			
			-- Create tile deck
			layerMod.deck = MOAIGfxQuad2D.new()
			layerMod.deck:setTexture( layerMod.image )
			layerMod.deck:setRect( x0, y0, x1, y1 )
			
			-- Create prop
			layerMod.prop = MOAIProp2D.new()
			layerMod.prop:setDeck( layerMod.deck )
			layerMod.prop:setPriority( layerMod.priority )
			layerMod.prop:setParent(tMap.prop)
			
			local xOff = layerMod.width / 2 - tMap.width / 2 + layerMod.x
			local yOff = tMap.height / 2 - layerMod.height / 2 - layerMod.y
			layerMod.prop:setLoc(xOff, yOff)
			
			rLayer:insertProp( layerMod.prop )
			
			if layerMod.modType == "replace" then
			
				-- Remove old prop
				layer.prop:setVisible(false)
				rLayer:removeProp(layer.prop)
				
				-- Update prop table
				local numLayerProps = #tMap.layerProps
				for j=1,numLayerProps do
					local layerProp = tMap.layerProps[j]
					if layerProp == layer.prop then
						tMap.layerProps[j] = layerMod.prop
						break
					end
				end
			end
		end
		
		tMap:updateLayerTransparencies()
	end
	
	function tMap:getProps(name)
		return tMap.MapData.props[name]
	end
	
	function tMap:getRandomProp(name)
		local props = tMap.MapData.props[name]
		local numProps = #props
		local propIndex = 0
		local numTries = 0
		local maxTries = 25
		while propIndex < 1 and numTries < maxTries do
			local index = math.random(1, numProps)
			-- Make sure it's empty
			if not props[index].rHero then
				propIndex = index
			else
				numTries = numTries + 1
			end
		end
		-- It's better to use the same prop rather then to block
		if numTries >= maxTries and propIndex < 1 then
			propIndex = math.random(1, numProps)
		end
		return props[propIndex]
	end
	
	function tMap:getRandomTile(name)
		local x = nil
		local y = nil
		local layout = tMap.MapData.layout	
		while not x and not y do
			local xi = math.random(1, layout.numTilesX)
			local yi = math.random(1, layout.numTilesY)
			local index = (yi - 1) * layout.numTilesX + xi
			local tileTypeIndex = layout.map[index]
			local tileType = layout.tileTypes[tileTypeIndex]
			if tileType.name == name then
				x = xi
				y = yi
			end
		end
		return x, y
	end
	
	function tMap:centerOnScreen(offsetX, offsetY, zoomOut, duration)
		Renderer.setCameraLoc(offsetX, offsetY, duration)
		if zoomOut then
			local scale = 2.0 * tMap.width / 640
			Renderer.setCameraScl(scale, scale, duration)
		else
			Renderer.setCameraScl(1, 1, duration)
		end
	end
	
	function tMap:enterOverviewMode()
		local offsetY = 0 -- tMap.height / 2.0
		tMap:centerOnScreen(0, -offsetY, true, 0.75)
	end
	
	function tMap:exitOverviewMode()
		tMap:centerOnScreen(0, 0, false, 0.75)
	end
	
	function tMap:centerOnTileType(sTileType, duration)
		local tileType = tMap:getTileTypeByName(sTileType)
		if tileType and tileType.centerTileX and tileType.centerTileY then
			local coords = tMap:getWorldCoords(tileType.centerTileX, tileType.centerTileY)
			print("Center on " .. sTileType .. ": " .. tostring(tileType.centerTileX) .. ", " .. tostring(tileType.centerTileY) .. " -> " .. tostring(coords.x) .. ", " .. tostring(coords.y))
			Renderer.setCameraLoc(coords.x, coords.y, duration)
		end
	end
	
	function tMap:updateLayerTransparencies()
		
		-- HACK HACK HACK
		local occupiedTiles = { }
		local numHeros = #g_tHeroes
		for i=1,numHeros do
			local hero = g_tHeroes[i]
			local tx, ty = hero:getCurrentTileCoords()
			table.insert(occupiedTiles, {tx, ty})
		end
		
		tMap:updateLayerTransparenciesForTiles(occupiedTiles)
	end
	
	function tMap:updateLayerTransparenciesForTiles(tOccupiedTiles)
	
		-- Fade in unoccupying layers
		local numLayers = #tMap.layerProps
		for i=1,numLayers do
			local prop = tMap.layerProps[i]
			prop:setColor(1, 1, 1, 1)
		end
	
		local trans = 0.5
		
		-- Fade out occupying layers
		local layout = tMap.MapData.layout	
		local numTiles = #tOccupiedTiles
		for i=1,numTiles do
			local layerIndex = numTiles - i + 1
			local tile = tOccupiedTiles[layerIndex]
			local index = (tile[2] - 1) * layout.numTilesX + tile[1]
			local layerIndex = layout.transparency[index]
			if layerIndex > 0 then
				local prop = tMap.layerProps[layerIndex]
				prop:setColor(trans, trans, trans, trans)
			end
		end
	end
	
	return tMap
end