local TerritoryManager = {}
local Territory = {}
local util = require('DFUtil')
local RobberyEvent = require('event/RobberyEvent')
local MissionDefs = require('mission/MissionDefs')
local Mission = require('mission/Mission')
local DFFile = require('DFCommon.File')
local DFUtil = require('DFCommon.Util')
local DFAnalytics = require('DFCommon.Analytics')

TerritoryManager.tTerritoryDefs = dofile(DFFile.getDataPath("Territories/TerritoryDefs.lua"))
function TerritoryManager.getTerritoryDef(sTerritoryKey)
    if sTerritoryKey then
        return TerritoryManager.tTerritoryDefs[sTerritoryKey]
    end
end
function TerritoryManager.updateTerritoryDef(sTerritoryKey, tValues)
    g_Game:updateDefWithTweak(TerritoryManager.tTerritoryDefs, sTerritoryKey, tValues)
end

TerritoryManager.tTerritoryTweakables = -- defaults
{
    nMaxConcurEvents = 2,
    tEventGenTime = 
    {   
        { nMax = 0, nTime = 60 },
        { nMax = 2, nTime = 160 },
        { nMax = 3, nTime = 400 },
        { nMax = 99, nTime = 500 },
    },
    nTerritoryRewardGenTime = 120,
    nMaxTotalEventsOnMap = 7,
    nTimeUntilHappIncrease = 120,
}

TerritoryManager.tTerritoryUnlockInfos =
{
    { "PummelPark" },
    { "StabbersRange" },
    { "BloodBeach" },
    { "BlightHeights" },
    { "BatteryRow" },
    { "DeadEndDrag" },   
    { "WarehouseDistrict" },
    { "DangerousDocks" },
    { "WestDump" },
    { "EastDump" },
}

TerritoryManager.tHappinessInfo =
{
    { sIcon = 'morale_superhigh', nRewardAmount = 30 },
    { sIcon = 'morale_high', nRewardAmount = 20 },
    { sIcon = 'morale_med', nRewardAmount = 15 },
    { sIcon = 'morale_low', nRewardAmount = 10 },
    { sIcon = 'morale_superlow', nRewardAmount = 5 },
}

function TerritoryManager.updateTerritoryTweakables(sDefKey, tValues)
    g_Game:updateDefWithTweak(TerritoryManager.tTerritoryTweakables, sDefKey, tValues)
end

function TerritoryManager.updateHappinessInfo(tInfo)
    if tInfo then
        TerritoryManager.tHappinessInfo = DFUtil.deepCopy(tInfo)
    end
end

function Territory:new(sDefKey, tSaved)
    local Ob = {}

    function Ob:getDef()
        return TerritoryManager.getTerritoryDef(self.sDefKey)
    end
    
    function Ob:addEvent(sDefKey, tSaved)
        local rDef = g_eventManager:getEventDef(sDefKey)
        if not rDef then
            print("ERROR: NULL def passed to addEvent")
            return nil
        end
        if rDef.bUseRandomOverride then
            local tOverrides = g_eventManager:getRandomEventOverride()
            if tOverrides then
                local rOverride = tOverrides[math.random(table.getn(tOverrides))]
                if rOverride then
                    for key, value in pairs(rOverride) do
                        rDef[key] = DFUtil.deepCopy(value)
                    end
                end
            end
        end
        if not tSaved then
            local tEventSaveTable = {}
            tEventSaveTable.sEventDefKey = sDefKey
            tEventSaveTable.tSaved = {}
            tSaved = tEventSaveTable.tSaved
            table.insert(self.tSaved.tEventSaveTables, tEventSaveTable) 
        end
        if rDef.eventClassName then
            local constructorFncn = g_eventManager:getConstructorFncn(rDef.eventClassName)
             if constructorFncn then
                local rNewEvent = constructorFncn(nil, self, rDef, tSaved)
                rNewEvent.sDefKey = sDefKey
                self.sLastAddedEventKey = sDefKey
                table.insert(self.tEvents, rNewEvent)
                if rNewEvent:isMissionEvent() then
                    self.rMissionEvent = rNewEvent
                end
                return rNewEvent
            end
        end        
        return nil
    end

    Ob.tSaved = tSaved
    Ob.tEvents = {}
    if not Ob.tSaved.tEventSaveTables then
        Ob.tSaved.tEventSaveTables = {}
    else
        -- add events
        for i, tEventSaveTable in ipairs(Ob.tSaved.tEventSaveTables) do
            Ob:addEvent(tEventSaveTable.sEventDefKey, tEventSaveTable.tSaved)
        end
    end

    Ob.bCleared = false
    Ob.nUnlockLevel = 0
    Ob.sDefKey = sDefKey   

    if not Ob.tSaved.tUnlockedMissionKeys then
        Ob.tSaved.tUnlockedMissionKeys = {}
    end

    if not Ob.tSaved.nCurrCrimeLevel then
        Ob.tSaved.nCurrCrimeLevel = Ob:getDef().nMaxCrimeLevel
    end

    if not Ob.tSaved.nTotPPRandomlyGenerated then
        Ob.tSaved.nTotPPRandomlyGenerated = 0
    end

    if not Ob.tSaved.nHappinessLevel then -- 1 is happy, 5 is unhappy        
        Ob.tSaved.nHappinessLevel = 1
    end

    if not Ob.tSaved.nSavedRewardAmount then
        Ob.tSaved.nSavedRewardAmount = 0
    end

    if not Ob.tSaved.nElapsedTimeWithNoEvents then
        Ob.tSaved.nElapsedTimeWithNoEvents = 0
    end

    Ob.sEventKeyOverride = nil
    Ob.rOverrideEvent = nil

    function Ob:setEventToGenerateOverride(sEventKey)
        if sEventKey then
            self.sEventKeyOverride = sEventKey
        end
    end

    function Ob:_getMissionDefKeyToUnlock()
        if self:getDef().tMissionInfos then
            local rDef = nil
            for i, rMissionInfo in ipairs(self:getDef().tMissionInfos) do
                local nCrimeLevel = rMissionInfo.nCrimeLevelToTrigger
                local sEventKey = rMissionInfo.sEventKey
                if nCrimeLevel and sEventKey then
                    if self.tSaved.nCurrCrimeLevel <= nCrimeLevel and
                        not self.tSaved.tUnlockedMissionKeys[sEventKey] then
                        self.tSaved.tUnlockedMissionKeys[sEventKey] = true
                        DFAnalytics.logEvent("prg_unlk_miss_" .. sEventKey)
                        return sEventKey
                    end
                end
            end
        end
        return rDef
    end

    function Ob:canRandomlyGeneratePP()        
        if self:getDef().tRewardOverride and self.tSaved.nTotPPRandomlyGenerated <= self:getDef().nMaxPP then
            local nPPChance = self:getDef().nPPChance
            if nPPChance then
                if math.random(1, 100) <= nPPChance then
                    return true
                end
            end
        end
        return false
    end

    function Ob:onRandomPPRewardGiven(nPPAmount)
        self.tSaved.nTotPPRandomlyGenerated = self.tSaved.nTotPPRandomlyGenerated + nPPAmount
    end

    function Ob:generateEvents(nMaxOverride)
        if self.sEventKeyOverride or self.rOverrideEvent then
            if not self.rOverrideEvent then
                self.rOverrideEvent = self:addEvent(self.sEventKeyOverride)
                self.sEventKeyOverride = nil
                return true
            end
            return false
        end
        local nNumEvents = table.getn(self.tEvents)
        local nNumEventsToAdd = TerritoryManager.tTerritoryTweakables.nMaxConcurEvents - nNumEvents
        if nNumEventsToAdd <= 0 then
            return false
        end
        local kMAX_EVENTS_TO_GENERATE = nMaxOverride or 2
        if nNumEventsToAdd > kMAX_EVENTS_TO_GENERATE then
            nNumEventsToAdd = kMAX_EVENTS_TO_GENERATE
        end
        for i = 1, nNumEventsToAdd do
            -- see if we can generate a mission event
            local bEventAdded = false
            if not bEventAdded then
                local sMissionEventKey = self:_getMissionDefKeyToUnlock()
                if sMissionEventKey then
                    local rNewEvent = self:addEvent(sMissionEventKey)                
                    if rNewEvent then
                        self.rMissionEvent = rNewEvent
                        rNewEvent:setIsMissionEvent(true)
                        bEventAdded = true
                    end
                end
            end

            if not bEventAdded then
                local sRareEventKey = g_eventManager:getRareEventKey()
                if sRareEventKey then
                    local rNewEvent = self:addEvent(sRareEventKey)
                    if rNewEvent then
                        bEventAdded = true
                    end
                end
            end

            if not bEventAdded then
                local sHolidayEventKey = g_eventManager:getHolidayEventKey()
                if sHolidayEventKey then
                    local rNewEvent = self:addEvent(sHolidayEventKey)
                    if rNewEvent then
                        bEventAdded = true
                    end
                end
            end

            if g_MiddleManager:isAtEndGame() then
                local sDefKey = g_eventManager:getEndGameRandomEventDef()
                if sDefKey then
                    local rNewEvent = self:addEvent(sDefKey)
                    bEventAdded = true
                end
            end            

            if not bEventAdded then
                -- generate a normal event
                local sNormalDefKey = self:getRandomCrimeEventDefKey()
                if sNormalDefKey then
                    local rNewEvent = self:addEvent(sNormalDefKey)
                    if self:canRandomlyGeneratePP() then
                        rNewEvent:setRewardOverride(self:getDef().tRewardOverride)
                    end
                    bEventAdded = true
                end
            end

            if bEventAdded then
                nNumEventsToAdd = nNumEventsToAdd - 1
            end
        end
        return true
    end

    function Ob:getRandomCrimeEventDefKey()
        local tEventInfos = self:getDef().tEventInfos
        if tEventInfos then
            for i, rInfo in ipairs(tEventInfos) do
                local bCanAdd = true
                if rInfo.nMaxCrimeLevel then
                    if self.tSaved.nCurrCrimeLevel > rInfo.nMaxCrimeLevel then
                        bCanAdd = false
                    end
                end
                if bCanAdd then
                    local tEventChanceInfos =  DFUtil.deepCopy(rInfo.tInfos)
                    for i, rEventChanceInfo in ipairs(tEventChanceInfos) do
                        if rEventChanceInfo.sEventKey == self.sLastAddedEventKey then
                            table.remove(tEventChanceInfos, i)
                            break
                        end
                    end
                    local nChance = math.random(100)
                    local rLastEventChanceInfo = nil
                    for i, rEventChanceInfo in ipairs(tEventChanceInfos) do
                        rLastEventChanceInfo = rEventChanceInfo
                        if nChance < rEventChanceInfo.nPossibility then
                            return rEventChanceInfo.sEventKey
                        end
                    end
                    -- didn't add anything so return the last one
                    if rLastEventChanceInfo then
                        return rLastEventChanceInfo.sEventKey
                    end
                    break
                end
            end
        end
    end

    function Ob:hasQueuedEvents()
        if table.getn(self.tEvents) > 0 then
            return true
        end
        return false
    end

    function Ob:getNumEvents()
        return table.getn(self.tEvents)
    end

    function Ob:getQueuedEvents()
        return self.tEvents
    end

    function Ob:onEventAborted(rEvent)
        if rEvent then
            if rEvent == self.rMissionEvent then
                self.rMissionEvent = nil
            end
            if rEvent == self.rOverrideEvent then
                self.rOverrideEvent = nil
            end
            for i, rExistingEvent in ipairs(self.tEvents) do
                if rEvent == rExistingEvent then
                    table.remove(self.tEvents, i)
                    if self.tSaved.tEventSaveTables[i] then
                        table.remove(self.tSaved.tEventSaveTables, i)
                    else
                        print("ERROR:COULD NOT REMOVE THE SAVE TABLE")
                    end
                    return
                end
            end
        end
        print("ERROR: COULD NOT FIND THE EVENT TO ABORT")
    end

    function Ob:removeEvent(rEvent)
        if rEvent == self.rMissionEvent then
            self.rMissionEvent = nil                    
        end
        if rEvent == self.rOverrideEvent then
            self.rOverrideEvent = nil
        end
        for i, rQueuedEvent in ipairs(self.tEvents) do
            if rEvent == rQueuedEvent then
                table.remove(self.tEvents, i)
                if self.tSaved.tEventSaveTables[i] then
                    table.remove(self.tSaved.tEventSaveTables, i)
                else
                    print("ERROR:COULD NOT REMOVE THE SAVE TABLE")
                end
                g_TodoList:onEventRemoved(rEvent, self)
--                g_City:enableEventGeneration(true)
                break
            end
        end
    end

    function Ob:hasActiveMission()
        if self.rMissionEvent then
            return true
        else
            return false
        end
    end

    function Ob:setUnlocked(bUnlock)        
        self.bUnlocked = bUnlock
        if not bUnlock and not self.bCleared then
            self.tSaved.nCurrCrimeLevel = Ob:getDef().nMaxCrimeLevel
        end
    end

    function Ob:setUnlockLevel(nLevel)    
        self.nUnlockLevel = nLevel
    end

    function Ob:getUnlockLevel()
        return self.nUnlockLevel
    end

    function Ob:getUnlocked()
        return self.bUnlocked
    end

    function Ob:setEventSuccessful(rEvent)
    end

    function Ob:setEventFailed(rEvent)
    end

    function Ob:setCleared(bCleared)
        if bCleared and not self.bCleared then        
            self.bCleared = true            
            DFAnalytics.logEvent("prg_clr_" .. self.sDefKey)
            g_City:onTerritoryCleared(self)            
        end
    end

    function Ob:isCleared()
        return self.bCleared
    end

    function Ob:getCrimeLevel()
        return (self.tSaved.nCurrCrimeLevel / self:getDef().nMaxCrimeLevel) * 100
    end

    function Ob:getBossFightProgress()
        local nDiff = self:getDef().nMaxCrimeLevel - self.tSaved.nCurrCrimeLevel
        return (nDiff / (self:getDef().nMaxCrimeLevel - 1)) * 100
    end

    function Ob:addCrimeAmount(nAmount)
        g_Game:addCrimeLevelChange(-nAmount)
        self.tSaved.nCurrCrimeLevel = self.tSaved.nCurrCrimeLevel + nAmount
        if self.tSaved.nCurrCrimeLevel < 0 then
            self.tSaved.nCurrCrimeLevel = 0
        elseif self.tSaved.nCurrCrimeLevel > self:getDef().nMaxCrimeLevel then
            self.tSaved.nCurrCrimeLevel = self:getDef().nMaxCrimeLevel
        end
    end

    function Ob:getClearedXPReward()
        return self:getDef().nClearedXPReward
    end

    function Ob:getClearedCashReward()
        return self:getDef().nClearedCashReward
    end

    function Ob:getClearedPPReward()
        return self:getDef().nClearedPPReward
    end

    function Ob:getTerritoryKey()
        return self:getDef().index
    end

    function Ob:getMapUIInfo()
        return self:getDef().tMapUIInfo
    end

    function Ob:hasMissionsToUnlock()
        if self:getDef().tMissionInfos then
            for i, rMissionInfo in ipairs(self:getDef().tMissionInfos) do
                local nCrimeLevel = rMissionInfo.nCrimeLevelToTrigger
                if nCrimeLevel then
                    if self.tSaved.nCurrCrimeLevel >= nCrimeLevel then
                        return true
                    end
                end
            end
        end
        return false
    end

    function Ob:getTerritoryHappinessInfo()
        if not self.tSaved.nHappinessLevel then
            self.tSaved.nHappinessLevel = 1
        end
        local nNumItems = table.getn(TerritoryManager.tHappinessInfo)
        if self.tSaved.nHappinessLevel > nNumItems then
            self.tSaved.nHappinessLevel = nNumItems
        end
        return TerritoryManager.tHappinessInfo[self.tSaved.nHappinessLevel]
    end

    function Ob:getHappinessIconName()
        local rInfo = self:getTerritoryHappinessInfo()
        if rInfo then
            return rInfo.sIcon
        end
        return ''
    end

    function Ob:getHappinessRewardAmount()
        return self.tSaved.nSavedRewardAmount
    end

    function Ob:collectHappinessReward()
        local nAmountToAdd = self.tSaved.nSavedRewardAmount
        g_Game:addCash(nAmountToAdd)
        g_TodoList:onMapCoinCollected(self.tSaved.nSavedRewardAmount)
        self.tSaved.nSavedRewardAmount = 0
        self.tSaved.nTimeSinceLastRewardCollection = MOAISim:getDeviceTime()
        return nAmountToAdd
    end

    function Ob:generateHappinessReward()
        if self.tSaved.nSavedRewardAmount == 0 then
            local rInfo = self:getTerritoryHappinessInfo()
            if rInfo and rInfo.nRewardAmount then
                self.tSaved.nSavedRewardAmount = rInfo.nRewardAmount
            end
        end
    end

    function Ob:incrementHappinessLevel()
        self:adjustHappinessLevel(-1)
    end
    
    function Ob:decrementHappinessLevel()
        self:adjustHappinessLevel(1)        
    end

    function Ob:adjustHappinessLevel(nAmount)
        if not self.tSaved.nHappinessLevel then
            self.tSaved.nHappinessLevel = 1
        end
        if nAmount < 0 then
            self.tSaved.nHappinessLevel = self.tSaved.nHappinessLevel + nAmount
            if self.tSaved.nHappinessLevel < 1 then
                self.tSaved.nHappinessLevel = 1
            end
        else
            local nNumItems = table.getn(TerritoryManager.tHappinessInfo)
            self.tSaved.nHappinessLevel = self.tSaved.nHappinessLevel + nAmount
            if self.tSaved.nHappinessLevel > nNumItems then
                self.tSaved.nHappinessLevel = nNumItems
            end
        end
    end

    function Ob:onTick(nElapsedTime)
        for i, rEvent in ipairs(self.tEvents) do
            rEvent:onTick(nElapsedTime)
        end
        if table.getn(self.tEvents) == 0 then
            self.tSaved.nElapsedTimeWithNoEvents = self.tSaved.nElapsedTimeWithNoEvents + nElapsedTime
            if self.tSaved.nElapsedTimeWithNoEvents >= TerritoryManager.tTerritoryTweakables.nTimeUntilHappIncrease then
                self:incrementHappinessLevel()
                self.tSaved.nElapsedTimeWithNoEvents = 0
            end
        end
        local nCurTime = MOAISim:getDeviceTime()
        if not self.tSaved.nTimeSinceLastRewardCollection then
            self.tSaved.nTimeSinceLastRewardCollection = nCurTime
        end
        if math.abs(self.tSaved.nTimeSinceLastRewardCollection - nCurTime) > TerritoryManager.tTerritoryTweakables.nTerritoryRewardGenTime then
            self:generateHappinessReward()
        end
    end

    function Ob:getTerritoryHappiness()
        return self.tSaved.nHappinessLevel
    end

    return Ob
end

function TerritoryManager:createCity()
	local Ob = {}
	
    local tSaveTable = g_Game:getSaveTable()
    if not tSaveTable.tTerritories then
        tSaveTable.tTerritories = {}
    end
	Ob.tTerritories = {}   
    for name, rDef in pairs(TerritoryManager.tTerritoryDefs) do
        if not tSaveTable.tTerritories[name] then
            tSaveTable.tTerritories[name] = {}
        end
        local rTerritory = Territory:new(name, tSaveTable.tTerritories[name])
        Ob.tTerritories[name] = rTerritory        
    end

    for nLevel, tLevelNames in ipairs(TerritoryManager.tTerritoryUnlockInfos) do
        for i, sLevelName in ipairs(tLevelNames) do
            if Ob.tTerritories[sLevelName] then
                Ob.tTerritories[sLevelName]:setUnlockLevel(nLevel)
            end
        end
    end    

    function Ob:getTerritory(sTerritoryName)
        if sTerritoryName then
            return self.tTerritories[sTerritoryName]
        end
        return nil
    end

	function Ob:getTerritories()
		return self.tTerritories
	end

    function Ob:getTerritoryBonus()
        local nBonusAmount = 0
        for i, rTerritory in pairs(self.tTerritories) do
            nBonusAmount = nBonusAmount + rTerritory:getTerritoryBonus()
        end
        return nBonusAmount
    end

    function Ob:getUnlockedTerritories()
        local tTerritories = {}
        for i, rTerritory in pairs(self.tTerritories) do
            if rTerritory:getUnlocked() then
                table.insert(tTerritories, rTerritory)
            end
        end
        return tTerritories
    end

    function Ob:getQueuedEvents(rHero) 
        -- we make the assumption of the hero being in one territory at a time
        for i, rTerritory in pairs(self.tTerritories) do
            if rTerritory:isHeroInTerritory(rHero) then
                return rTerritory:getQueuedEvents(rHero)
            end
        end
        return nil
    end

    function Ob:areEventsQueued()
        for i, rTerritory in pairs(self.tTerritories) do
            if rTerritory:hasQueuedEvents() then
                return true
            end
        end
        return false
    end
    
    function Ob:getNumRareEvents(sRareEventKey)
        local nNumEvents = 0
        for i, rTerritory in pairs(self.tTerritories) do
            local tEvents = rTerritory:getQueuedEvents()
            if tEvents then
                for i, rEvent in ipairs(tEvents) do
                    local rDef = rEvent:getDef()
                    if rDef and rDef.sRareEventKey and rDef.sRareEventKey == sRareEventKey then
                        nNumEvents = nNumEvents + 1
                    end
                end
            end
        end
        return nNumEvents
    end

    function Ob:setUnlockLevel(nLevel)
        self.nUnlockLevel = nLevel
        for i, rInfo in ipairs(TerritoryManager.tTerritoryUnlockInfos) do
            if i <= nLevel then
                for c, sTerritoryName in ipairs(rInfo) do
                    local rTerritory = self.tTerritories[sTerritoryName]
                    if rTerritory then
                        rTerritory:setUnlocked(true)
                    end
                end
            else
                -- this makes sure that the areas not unlocked are locked
                -- it also resets the crime level
                for c, sTerritoryName in ipairs(rInfo) do
                    local rTerritory = self.tTerritories[sTerritoryName]
                    if rTerritory then
                        rTerritory:setUnlocked(false)
                    end
                end                
            end
        end
    end

    function Ob:getUnlockLevel()
        return self.nUnlockLevel
    end

    function Ob:areAllTerritoriesUnlocked()
        local bAllUnlocked = false
        
        for i, rTerritory in pairs(self.tTerritories) do
            if rTerritory and true ~= rTerritory:getUnlocked() then
                bAllUnlocked = false
            end
        end
        
        return bAllUnlocked
    end
    
    function Ob:onTerritoryCleared(rTerritory)
        -- process through all the unlocks
        local tLevelInfos = TerritoryManager.tTerritoryUnlockInfos[self:getUnlockLevel()]
        local bAllCleared = true
        if tLevelInfos then
            for i, sTerritoryName in ipairs(tLevelInfos) do
                local rTerritory = self.tTerritories[sTerritoryName]
                if rTerritory then
                    if not rTerritory:isCleared() then
                        bAllCleared = false
                    end
                end
            end
        end
        if bAllCleared then
            g_MiddleManager:incrementLevel(1)
            self:setUnlockLevel(g_MiddleManager:getLevel())
        end
    end

    function Ob:enableEventGeneration(bEnable)
        self.bEventGenEnabled = bEnable
    end

    function Ob:forceGenerateEvents()
        self.nTimeSinceEventGeneration = 100000
    end

    function Ob:getNumEventsAvailable()
        local nNumEvents = 0
        for i, rTerritory in pairs(self.tTerritories) do
            nNumEvents = nNumEvents + rTerritory:getNumEvents()
        end
        return nNumEvents
    end

    function Ob:getEventGenTime(nNumAvailableEvents)        
        local nEventGenTime = 0
        if TerritoryManager.tTerritoryTweakables.tEventGenTime and nNumAvailableEvents then
            for i, rGenTimeInfo in ipairs(TerritoryManager.tTerritoryTweakables.tEventGenTime) do
                if nNumAvailableEvents > rGenTimeInfo.nMax then
                    nEventGenTime = rGenTimeInfo.nTime
                end
            end
        end
        return nEventGenTime
    end

    function Ob:getActiveEventsInSet(sSetName)
        local nNumActiveEvents = 0
        for i, rTerritory in pairs(self.tTerritories) do
            local tEvents = rTerritory:getQueuedEvents()
            if tEvents then
                for i, rEvent in ipairs(tEvents) do
                    local rDef = rEvent:getDef()
                    if rDef then
                        if rDef.sEventSetName == sSetName then
                            nNumActiveEvents = nNumActiveEvents + 1
                        end
                    end
                end
            end
        end
        return nNumActiveEvents
    end

    function Ob:onTick(nElapsedTime)
        if self.bEventGenEnabled then
            self.nTimeSinceEventGeneration = self.nTimeSinceEventGeneration + nElapsedTime
            local nNumAvailableEvents = self:getNumEventsAvailable()
            local nEventGenTime = self:getEventGenTime(nNumAvailableEvents)
            if self.nTimeSinceEventGeneration >= nEventGenTime then
                if nNumAvailableEvents >= TerritoryManager.tTerritoryTweakables.nMaxTotalEventsOnMap then
                    return
                end
                local rTerritoryToGenEvent = nil
                local nCurLevel = g_MiddleManager:getLevel()
                if nCurLevel then
                    local sTerritoryName = TerritoryManager.tTerritoryUnlockInfos[nCurLevel][1]
                    if sTerritoryName then
                        local rTerritory = self.tTerritories[sTerritoryName]
                        if rTerritory then
                            if rTerritory:getNumEvents() < TerritoryManager.tTerritoryTweakables.nMaxConcurEvents then
                                rTerritoryToGenEvent = rTerritory
                            end
                        end
                    end
                end

                if not rTerritoryToGenEvent then
                    local tTerritories = self:getUnlockedTerritories()
                    local tPossibleTerritories = {}
                    for i, rTerritory in ipairs(tTerritories) do
                        if rTerritory:getNumEvents() < TerritoryManager.tTerritoryTweakables.nMaxConcurEvents then
                            table.insert(tPossibleTerritories, rTerritory)
                        end
                    end
                    
                    local nNumTerritories = table.getn(tPossibleTerritories)
                    if nNumTerritories > 0 then
                        local nRandNum = math.random(1, nNumTerritories)
                        rTerritoryToGenEvent = tPossibleTerritories[nRandNum]                        
                    end
                end

                if rTerritoryToGenEvent then
                    rTerritoryToGenEvent:generateEvents(1)
                end
                self.nTimeSinceEventGeneration = 0
            end
        end
        for i, rTerritory in pairs(self.tTerritories) do
            if rTerritory:getUnlocked() then
                rTerritory:onTick(nElapsedTime)
            end
        end
    end
    
    Ob:setUnlockLevel(g_MiddleManager:getLevel())
    Ob.nTimeSinceEventGeneration = 0

	return Ob
end
    
return TerritoryManager
