module(..., package.seeall)

local Renderer = require('Renderer')
local util = require("DFUtil")

kTEXTEFFECTINSTANCE_SLEEP_TIME = 0.025

function TextEffectInstance:new(effectDef, value, renderLayer, x, y, priority)
	-- CONSTRUCTOR
	local rEffect = nil
	if effectDef then
		rEffect = util.dofile(effectDef)
	end
	
	if not rEffect then
		print("Unable to load text effect definition: " .. effectDef)
		return nil
	end
	
	rEffect.activeInstances = { }
	rEffect.value = value
	rEffect.renderLayer = renderLayer
	rEffect.x = x
	rEffect.y = y
	rEffect.priority = priority
	rEffect.isRunning = false
	
	function rEffect:getRandom(scale)
		if not scale then scale = 1 end
		return math.random(0, 10000) / 10000.0
	end
	
	function rEffect:getCenteredRandom(scale)
		if not scale then scale = 1 end
		return scale * math.random(-10000, 10000) / 10000.0
	end
	
	function rEffect:Execute()
		
		local tLayer = Renderer.getLayer(rEffect.renderLayer)
		
		local t = 0
		local lastT = 0
		local isDone = false
		
		local spawnCooldown = 0
		
		while not isDone do

			local dt = t - lastT
			
			local isSpawning = t < rEffect.spawnDuration
			if isSpawning then
				if spawnCooldown <= 0 then
					spawnCooldown = 1 / rEffect.spawnRate
					
					-- Spawn a new text effect
					local textBox = g_GUI:CreateTextBox(rEffect.pointSize, MOAITextBox.CENTER_JUSTIFY, rEffect.font)
					
					table.insert(rEffect.activeInstances, textBox)
					
					-- General setup
					textBox:setRect(-100, -100, 100, 100)
					textBox:setPriority(rEffect.priority)					
					textBox:setString(rEffect.value)
					
					-- Position
					local a = rEffect:getCenteredRandom(0.5)
					local b = rEffect:getCenteredRandom(0.5)
					local x = rEffect.x + a * rEffect.emitterWidth + rEffect.offsetX
					local y = rEffect.y + a * rEffect.emitterHeight + rEffect.offsetY
					textBox:setLoc(x, y)
					
					-- Timing
					a = rEffect:getRandom()
					textBox.age = (1 - a) * rEffect.minAge + a * rEffect.maxAge
					textBox.endTime = t + textBox.age
					
					-- Movement
					local dx = rEffect.dirX
					local dy = rEffect.dirY
					local mag = math.sqrt(dx * dx + dy *dy)
					dx = dx / mag
					dy = dy / mag
					
					a = rEffect:getRandom()
					local speed = (1 - a) * rEffect.minSpeed + a * rEffect.maxSpeed
					
					local moveAge = textBox.age + 1
					dx = dx * speed * moveAge
					dy = dy * speed * moveAge
					textBox:moveLoc(dx, dy, moveAge, MOAIEaseType.LINEAR)
					
					-- Color
					local dc = {}
					for i=1,4 do
						dc[i] = rEffect.endColor[i] - rEffect.startColor[i]
					end
					textBox:setColor(rEffect.startColor[1], rEffect.startColor[2], rEffect.startColor[3], rEffect.startColor[4])
					textBox:moveColor(dc[1], dc[2], dc[3], dc[4], textBox.age, MOAIEaseType.LINEAR)
					
					-- Add to renderer
					tLayer.RenderLayer:insertProp(textBox)
				else
					spawnCooldown = spawnCooldown - dt
				end
			end
			
			-- Update the instances
			isDone = not isSpawning
			
			local numInstances = #rEffect.activeInstances
			for i=1,numInstances do
				local index = numInstances - i + 1
				local instance = rEffect.activeInstances[index]
				if t >= instance.endTime then
					tLayer.RenderLayer:removeProp(instance)
					table.remove(rEffect.activeInstances, index)
				else
					isDone = false
				end
			end
			
			-- Yield for other threads
			util.sleep(kTEXTEFFECTINSTANCE_SLEEP_TIME)
			
			-- Update timer
			lastT = t
			t = t + kTEXTEFFECTINSTANCE_SLEEP_TIME
		end
	
		self.rThread = nil
		self.isRunning = false
	end
	
	function rEffect:start()
		self.rThread = MOAIThread.new()
		self.rThread:run(self.Execute, self)
		self.isRunning = true
	end
	
	function rEffect:stop()
		if self.rThread then
			self.rThread:stop()
		end
	end
	
	function rEffect:isDone()
		return not self.isRunning
	end
	
	return rEffect
end