local ResearchActivity = {}

local DFAnalytics = require('DFCommon.Analytics')
local util = require('DFUtil')
local Activity = require('activity/Activity')
local WalkToRoomTask = require('WalkToRoomTask')
local WaitForPropTask = require('WaitForPropTask')
local Item = require('Item')
local TileMap = require('TileMap')
local Renderer = require('Renderer')
local CountdownUI = require('ui/CountdownUI')
local DFAchievement = require('DFCommon.Achievement')

function ResearchActivity:new(rHero, rDef, nIndex, bRunUntilSuccess, tSaved, sDefKey)
	local Ob = Activity:new(rHero, rDef, nIndex, bRunUntilSuccess, tSaved)
    Ob.tSaved = tSaved
    if sDefKey then
        Ob.tSaved.sDefKey = sDefKey
    end

	function Ob:Run()
        local rOverlayLayer = Renderer.getRenderLayer('GameplayOverlay')
        self.rCountdownUI = CountdownUI:new(rOverlayLayer)
        self.rCountdownUI:setLoc(0, 200)
        self.rCountdownUI:setVisible(true)
        self.rCountdownUI:setParent(self.rHero)

        self.rProp = self.rAvailableProp
        if not self.tSaved.sDefKey or not self.rProp then
            self:EndRun()
            return
        end

        self.rHero:setPriority(self.rProp:getInteractPri())
		local orgX, orgY = Ob.rHero:getLoc()
        local nInteractOffsetX, nInteractOffsetY = self.rProp:getInteractOffset()
		self.rHero:setLoc(orgX + nInteractOffsetX, orgY + nInteractOffsetY)
        local sInteractAnimName, bFlipX = self.rProp:getInteractAnim()
        self.rHero:setFacingDirection((bFlipX and "SE") or "SW")	
		self.rHero:playAnimation(sInteractAnimName, false, bFlipX)

        local bIsMeteorite = false
        local rDef = Item.getEquipmentDef(self.tSaved.sDefKey)
        if not rDef then
            rDef = Item.getConsumableDef(self.tSaved.sDefKey)
        end
        if not rDef then
            rDef = Item.getMeteoriteDef(self.tSaved.sDefKey)
            bIsMeteorite = true
        end

        if not rDef then
            self:EndRun()
            return            
        end

        self.nDuration = rDef.nResearchDuration or 1
        local nPropMult = self.rProp:getPropDef().researchSpeed or 1
        -- apply the morale mult
        self.nDuration = self.nDuration / nPropMult
        self.nDuration = self.nDuration * (1 / g_Game:getDebugTimeScale())
        local nSpeedMod = self.rHero:getActivitySpeedMod(self.rDef.sActivityType)
        if nSpeedMod > 0 then
            self.nDuration = self.nDuration * (1 - (math.min(100, nSpeedMod) / 100))
        end

        self.nElapsedTime = 0
        self.bDone = false

        local nLevel = self.rProp:getPropDef().nLevel or 1
        self.nElapsedTime = g_MiddleManager:getResearchProgress(self.tSaved.sDefKey) * self.nDuration

        while not self.bDone do
            local nCurTime = MOAISim:getDeviceTime()
            if not self.tSaved.nLastTimeChange or self.tSaved.bWasForceKilled then
                self.tSaved.bWasForceKilled = false
                self.tSaved.nLastTimeChange = nCurTime
            end
            local nChange = math.abs(nCurTime - self.tSaved.nLastTimeChange)
            self.nElapsedTime = self.nElapsedTime + nChange
            g_MiddleManager:incrementResearchProgress(self.tSaved.sDefKey, nChange / self.nDuration)

            if self.nElapsedTime >= self.nDuration then
                self.bDone = true
            end            

            self.tSaved.nLastTimeChange = nCurTime
            self.rCountdownUI:updateProgress(self:getActivityProgress())
            util.sleep(1)
        end
        
        DFAchievement.updateAchievement('ResearchOneItem', 1)
        DFAchievement.updateAchievement('ResearchAllItems', self.tSaved.sDefKey)
        if bIsMeteorite then
            DFAchievement.updateAchievement('ResearchOneRecipe', 1)
        end
        
        DFAnalytics.logEvent("act_rsch_complete", { hero = self.rHero:getCharDefKey(), subject = self.tSaved.sDefKey } )        

        self.rHero:setOverlayText("COMPLETED RESEARCH", 1, {0, 1, 0})

		self.rHero:setLoc(orgX, orgY)
		self.rHero:playAnimation("breathe")	

		self:EndRun()	
	end

    function Ob:getRemainingTime()
        if not self.nDuration or not self.nElapsedTime then
            return 0
        end
        return math.max((self.nDuration - self.nElapsedTime), 0)
    end

    function Ob:getActivityProgress()
        if not self.nDuration or self.nDuration == 0 then
            return 100
        end
        return (((self.nElapsedTime or 0) / self.nDuration) * 100)
    end

    function Ob:getRushPrice()
        return self:computeRushCost(self:getRemainingTime())
    end

    function Ob:rushActivity()
        self.bDone = true
        g_MiddleManager:incrementResearchProgress(self.tSaved.sDefKey, 1)
        
        DFAnalytics.logEvent("act_rshd", { act = "rsch", remaining = self:getRemainingTime() } )
        DFAnalytics.logEvent("act_rsch_rshd", { hero = self.rHero:getCharDefKey(), cost = self:getRushPrice() } )
        DFAnalytics.incrStat("act_rsch_rshd", 1)        

        g_Audio:playCue('MMOJ_Audio/UI/RushTraining')
    end

	function Ob:OnStart()
		local propName = "research"
        local rRoom = g_CurrentMap:getRoom("lab")
        if not rRoom then
            -- exit
        end
        g_MiddleManager:setIsResearching(self.tSaved.sDefKey, true)

        local bTeleportToProp = false

		local task = nil
        if not self.tSaved.bInRoom then            
            task = WalkToRoomTask:new(self, 'lab')
            self:AddTask(task)
        else
            bTeleportToProp = true
        end
        self.tSaved.bInRoom = true   
   
        if not self.tSaved.bStarted then
            DFAnalytics.logEvent("act_research", { hero = self.rHero:getCharDefKey(), subject = self.tSaved.sDefKey  } )
            DFAnalytics.incrStat("act_research", 1)
            self.tSaved.bStarted = true
        end   
       
        task = WaitForPropTask:new(self, rRoom, propName, self.rHero, bTeleportToProp)
		self:AddTask(task)
	end

	function Ob:OnEnd(bForceKilled)
        if self.rAvailableProp then
            self.rAvailableProp:removeUser(self.rHero)
        end
        if not self.bDone and bForceKilled then
            self.tSaved.bInRoom = false
            self.tSaved.bWasForceKilled = true
        else
            self.tSaved.bWasForceKilled = false
        end
        if self.rCountdownUI then
            self.rCountdownUI:setVisible(false)
        end
        g_MiddleManager:setIsResearching(self.tSaved.sDefKey, false)
        
        self.tSaved.bStarted = false
	end

	return Ob
end

return ResearchActivity
