local Event = {}
local util = require('DFUtil')
local DFFile = require('DFCommon.File')
local DFUtil = require('DFCommon.Util')
local DFGraphics = require('DFCommon.Graphics')
local DFAnalytics = require('DFCommon.Analytics')
local EventActor = require('event/EventActor')
local EffectInstance = require('EffectInstance')
local Character = require('Character')
local Renderer = require('Renderer')
local SpriteSheet = require('SpriteSheet')
local DFAchievement = require('DFCommon.Achievement')
local AutoResolveEventActor = require('event/AutoResolveEventActor')
local Item = require('Item')

local EventBG = {}
local BGSlot = {}
local Meteorite = {}

Event.eRelDifficultyLevel =
{
    DL_EASY = 1,
    DL_MEDIUM = 2,
    DL_HARD = 3,
}

Event.tMiscTweaks =
{
    tStarRatingInSec =
    {
        480, -- 1 star
        360, -- 2 stars
        240, -- 3 stars
        120, -- 4 stars
    },
    kAUTO_RESOLVE_TIME = 60,
    nAutoResolvePlayerDamageFactor = 1,
    nAutoResolveEnemyDamageFactor = 1,
    nAutoResolvePlayerAdditiveDamage = 0,
    nAutoResolveEnemyAdditiveDamage = 0,
}

function Event.updateMiscTweakables(sDefKey, tValues)
    g_Game:updateDefWithTweak(Event.tMiscTweaks, sDefKey, tValues)
end

function Event.getMaxStarRating()
    return table.getn(Event.tMiscTweaks.tStarRatingInSec)
end

Event.tStaticTextures = nil

local kTIME_RATING_SCORE_CONV = 10
function Event.getTimeStarScoreMult()
    return kTIME_RATING_SCORE_CONV
end

local kINIT_FIGHT_X = -100
local kINIT_FIGHT_Y = 0
local kFIGHT_X = -44
local kFIGHT_Y = 0
local kCHARGEBAR_SCALE = 30
local kDEFAULT_AITRIGGER_DELAY = 3
local kMAX_TIME = 3600

Event.nDelayForDefaultAbilityCue = 0.3

function BGSlot:new(tLoc)
    Ob = {}
    Ob.x = tLoc[1]
    Ob.y = tLoc[2]
    Ob.rOccupant = nil

    function Ob:getLoc()
        return self.x, self.y
    end

    function Ob:getOccupant()
        return self.rOccupant
    end
    
    function Ob:setOccupant(rOccupant)
        self.rOccupant = rOccupant
    end

    return Ob
end

function EventBG:new(rDef, rRenderLayer)
	local Ob = MOAITransform.new()

    Ob.rDef = rDef
    Ob.rRenderLayer = rRenderLayer
    Ob.tOccupantToSlot = {}

    Ob.tBGPropTexturePaths = {}
    Ob.tBGPropProps = {}

    function Ob:getAmbienceCue()
        return self.rDef.sAmbienceCue
    end

    function Ob:initializeSlots(tSlots, tLocs)
        for i, tLoc in ipairs(tLocs) do
            local rNewSlot = BGSlot:new(tLoc)
            table.insert(tSlots, rNewSlot)
        end
    end

    function Ob:getMeteoritePos()
        if self.rDef.tSlots.MeteoriteLocs then
            local nNum = table.getn(self.rDef.tSlots.MeteoriteLocs)
            if nNum > 0 then
                return self.rDef.tSlots.MeteoriteLocs[math.random(nNum)]
            end
        end
        return { 0, 0 }
    end

    -- initialize slots
    for slotName, tSlotInfo in pairs(rDef.tSlots) do
        local tSlots = {}
        Ob:initializeSlots(tSlots, tSlotInfo)
        Ob["t"..slotName] = tSlots
    end

    function Ob:getNextSlot(slotName)
        local tSlots = self["t"..slotName]
        if tSlots then
            for i, rSlot in ipairs(tSlots) do
                if not rSlot:getOccupant() then
                    return rSlot
                end
            end
        end
        return nil
    end

    function Ob:reserveNextSlot(slotName, rOccupant)
        local rSlot = self:getNextSlot(slotName)
        if rSlot then
            -- unreserve previous slot
            local rPrevSlot = self.tOccupantToSlot[rOccupant]
            if rPrevSlot then
                rPrevSlot:setOccupant(nil)
                self.tOccupantToSlot[rOccupant] = nil
            end            
            rSlot:setOccupant(rOccupant)
            self.tOccupantToSlot[rOccupant] = rSlot
            return rSlot
        end
        return false
    end

    function Ob:createBGSprite()
        if self.rDef.eventSpriteSheet and self.rDef.spriteSheetIndex then
            local rSprite = SpriteSheet:new(self.rDef.eventSpriteSheet, nil, nil, nil, 'center', 'center')
            if rSprite then
                rSprite:setSpriteIndex(self.rDef.spriteSheetIndex)
                return rSprite
            end
        end
        return nil
    end

    function Ob:setVisible(bVisible)
        if not self.rRenderLayer then
            return
        end
        if bVisible then
            self.rBackdrop = self:createBGSprite()
            if self.rBackdrop then
                self.rBackdrop:setLoc(0, 0)
                self.rRenderLayer:insertProp(self.rBackdrop)
                self.rBackdrop:setParent(self)
            end
            for i, rProp in ipairs(self.tBGPropProps) do
                self.rRenderLayer:insertProp(rProp)
                rProp:setParent(self)
            end
        else
            if self.rBackdrop then
                self.rRenderLayer:removeProp(self.rBackdrop)
                self.rBackdrop:unloadSpriteSheet()
            end
            for i, rProp in ipairs(self.tBGPropProps) do
                self.rRenderLayer:removeProp(rProp)
                rProp:unloadSpriteSheet()
            end
            self.tBGPropProps = {}
        end
    end

    function Ob:addBGProp(sSpriteSheetPath, sSpriteName, x, y)
        if sSpriteSheetPath and sSpriteName then
            local rBGSprite = SpriteSheet:new(sSpriteSheetPath, nil, nil, nil,'center', 'center')
            if rBGSprite then
                rBGSprite:setSpriteIndex(sSpriteName)
                rBGSprite:setLoc(x, y)
                self.rRenderLayer:insertProp(rBGSprite)
                table.insert(self.tBGPropProps, rBGSprite)
                rBGSprite:setParent(self)
            end
        end
    end

    return Ob
end

function Meteorite:new(rMeteoriteInfo)
    local Ob = SpriteSheet:new('UI/hud', 'Characters/AnimDefs/Meteorite.lua', 0, 0, 'left', 'top')

    function Ob:moveThread()
        local bUp = true
        self.nOrigX, self.nOrigY = self:getLoc()
        local nCurYDelta = 0
        local nCurFrame = 0
        
        local speed = 1 / 30
        local moveRange = 3.5
        
        while not self.bStop do
            --print("self.bStop: " .. tostring(self.bStop))    
    
            nCurFrame = nCurFrame + 1
            nCurYDelta = math.sin(nCurFrame * speed) * moveRange
            
            self:setLoc(self.nOrigX, self.nOrigY + nCurYDelta)
            --self:seekScl(1, 1, 1, 0.5, MOAIEaseType.LINEAR)                
            
            coroutine.yield()
        end
    end

    function Ob:insertIntoLayer(tPos)
        local rRenderLayer = Renderer.getRenderLayer('EventFX')
        if rRenderLayer then
            rRenderLayer:insertProp(self)
        end
        self:playAnimation('Idle')
        if tPos then
            self:setLoc(tPos[1], tPos[2])
        end
        self.bStop = false
        self.rMoveThread = MOAIThread.new()
        self.rMoveThread:run(self.moveThread, self)
    end
    
    function Ob:removeFromLayer()
        local rRenderLayer = Renderer.getRenderLayer('EventFX')
        if rRenderLayer then
            rRenderLayer:removeProp(self)
        end
        self.bStop = true
        self.rMoveThread = nil -- stop the thread?
    end

    function Ob:onFinger(eventType, sx, sy, tProps)
        if eventType == MOAITouchSensor.TOUCH_UP then
            if tProps then
                for i, rProp in ipairs(tProps) do
                    if rProp == self then
                        self:open()
                    end
                end
            end
        end        
    end

    function Ob:open()
        if not self.rMeteoriteInfo.bTapped then
            self.rMeteoriteInfo.bTapped = true
            if self.rMeteoriteInfo.sKey and self.rMeteoriteInfo.nLevel then
                g_Audio:playCue('MMOJ_Audio/SFX/Battle/CollectMeteorite')
                local rDef = Item.getMeteoriteDef(self.rMeteoriteInfo.sKey)
                self:playAnimation(tostring(self.rMeteoriteInfo.nLevel))
                if rDef and rDef.tColor then
                    self:setColor(rDef.tColor[1], rDef.tColor[2], rDef.tColor[3])
                end
                local rEffectInstance = EffectInstance:new('Effects/MeteoriteTap.lua', Renderer.getRenderLayer('EventFX'))
                
                local function playEffect()                
                    local x, y = self.rMeteoriteInfo.rMeteor:getLoc()
                    if rEffectInstance then
                        rEffectInstance:setLoc(x+70, y-50)
                        rEffectInstance:setScl(3,3)
                        rEffectInstance:start()
                    end
                    
                    self.bStop = true
                    self:moveColor(0, 0, 0, -1, 2)
                    util.sleep(2)
                    self:removeFromLayer()
                end
                
                local rEffectThread = MOAIThread.new()
                rEffectThread:run(playEffect, self)
            end
        end
    end

    Ob.bOpened = false
    Ob.rMeteoriteInfo = rMeteoriteInfo    
    Ob.bStop = false

    return Ob
end

function Event:new(rTerritory, rDef, tSaved)
    local Ob = MOAITransform:new()
    Ob.__typename = "Event"

    Ob.tHeroes = {}
    Ob.tSaved = tSaved
	Ob.rLayer = nil
	Ob.rDef = rDef
	Ob.bSuccessful = false
	Ob.rTerritory = rTerritory
    Ob.bClearedTerritory = false
    Ob.rEventBG = nil
    Ob.bFleeRequested = false
    Ob.tQueuedHeroes = {}
    Ob.bAllowAddHeroes = true
    Ob.sSpriteSheetKey = 'tFightEventSpriteSheetInfo'
    
    if not Ob.tSaved.sEventType then
        Ob.tSaved.sEventType = nil
    end
    if not Ob.tSaved.nElapsedTime then
        Ob.tSaved.nElapsedTime = 0
    end
    if not Ob.tSaved.nStarRating then
        Ob.tSaved.nStarRating = 4
    end
    Ob.bHasCompleted = false

    if not Ob.tSaved.bIsMissionEvent then
        Ob.tSaved.bIsMissionEvent = false
    end

    function Ob:getSpriteSheetKey()
        return self.sSpriteSheetKey
    end

    function Ob:getDef()
        return self.rDef
    end

    function Ob:getNumMaxHeroes()
        return self.rDef.nMaxHeroes or 4
    end

    function Ob:getEventType()
        if not self.tSaved.sEventType then
            self.tSaved.sEventType = self.rDef.sEventType
        end
        return self.tSaved.sEventType
    end

    function Ob:setEventType(sEventType)
        self.tSaved.sEventType = sEventType
    end

	function Ob:getDescription()
		return rDef.sName
	end

	function Ob:getLevel()
		return self.rDef.nLevel
	end

    function Ob:enableTimeIncrement(bEnable)
        self.tSaved.bIncrementElapsedTime = bEnable
    end

    function Ob:onTick(nElapsedTime)
        if self.tSaved.bIncrementElapsedTime then            
            self.tSaved.nElapsedTime = self.tSaved.nElapsedTime + nElapsedTime
            if self.tSaved.nElapsedTime > kMAX_TIME then
                self.tSaved.nElapsedTime = kMAX_TIME
            end
            local nCurRating = self.tSaved.nStarRating
            for nRating, nTimeMin in ipairs(Event.tMiscTweaks.tStarRatingInSec) do
                if self.tSaved.nElapsedTime < nTimeMin then
                    nCurRating = nRating
                end
            end            
            if nCurRating ~= self.tSaved.nStarRating then
                if self.rTerritory then
                    if nCurRating < self.tSaved.nStarRating then
                        self.rTerritory:decrementHappinessLevel()
                    end
                end
                self.tSaved.nStarRating = nCurRating
            end            
        end
    end

    function Ob:getElapsedTime()
        if not self.tSaved.nElapsedTime then
            self.tSaved.nElapsedTime = 0
        end
        return self.tSaved.nElapsedTime
    end

    function Ob:getTimeStarRating()
        if not self.tSaved.nStarRating then
            self.tSaved.nStarRating = 4
        end
        return self.tSaved.nStarRating
    end

    function Ob:getTotalRatingScore()
        local nScore = Event.getTimeStarScoreMult() * self:getTimeStarRating()
        return nScore
    end

    function Ob:getRemainingTimeAtCurStarLevel()
        if not self.tSaved.nElapsedTime then
            self.tSaved.nElapsedTime = 0
        end
        local nCurRating = 1
        local nRemainingTime = 0
        for nRating, nTimeMin in ipairs(Event.tMiscTweaks.tStarRatingInSec) do
            if self.tSaved.nElapsedTime < nTimeMin then
                nCurRating = nRating
                nRemainingTime = nTimeMin - self.tSaved.nElapsedTime
            end
        end
        return nRemainingTime, nCurRating        
    end

    function Ob:setRewardOverride(tOverride)
        self.tRewardOverride = tOverride
    end

    function Ob:getRewardOverride()
        return self.tRewardOverride
    end

	function Ob:getXPReward()
        if self.tRewardOverride then
            if self.tRewardOverride.WinXP then
                return self.tRewardOverride.WinXP
            end
        end
        return self.rDef.Rewards.WinXP
	end

	function Ob:getCashReward()
        if self.tRewardOverride and not self.tRewardOverride.Cash then
            return 0
        end
        if self.tRewardOverride and self.tRewardOverride.Cash then
            return self.tRewardOverride.Cash
        end
        if self.rDef.Rewards.Cash then
            return self.rDef.Rewards.Cash
        end
        return 0
	end

	function Ob:getPPReward()
        if self.tRewardOverride and not self.tRewardOverride.PP then
            return 0
        end
        if self.tRewardOverride and self.tRewardOverride.PP then
            return self.tRewardOverride.PP
        end
        if self.rDef.Rewards.PP then
            return self.rDef.Rewards.PP
        end
        return 0
	end

    function Ob:getConsumableReward()
        if self.rDef.Rewards.sConsumableDef then
            return self.rDef.Rewards.sConsumableDef
        end
    end

    function Ob:getEquipmentDropReward()
        if self.rDef.Rewards.tDrops then
            local nMaxIndex = table.getn(self.rDef.Rewards.tDrops)
            if nMaxIndex > 0 then
                return self.rDef.Rewards.tDrops[math.random(nMaxIndex)]
            end
        end
        return nil
    end

    function Ob:addHeroToEvent(rHero)
        table.insert(self.tHeroes, rHero)
        rHero:onAssignedToEvent(self)
    end

    function Ob:removeHeroFromEvent(rHero)
        for i, rExitingHero in ipairs(self.tHeroes) do
            if rHero == rExitingHero then
                rHero:onAssignedToEvent(nil)
                table.remove(self.tHeroes, i)
            end
        end
    end

    function Ob:clearAllHeroes()
        for i, rHero in ipairs(self.tHeroes) do
            rHero:onAssignedToEvent(nil)
        end
        self.tHeroes = {}
    end

    function Ob:hasHeroesAssigned()
        if self.tHeroes and table.getn(self.tHeroes) > 0 then
            return true
        end
        return false
    end

    function Ob:getAllHeroes()
        return self.tHeroes
    end

    function Ob:getAllEventEnemies()        
        return self.tEnemyActors
    end

    function Ob:getAllHeroActors()
        local tHeroActors = {}
        local nHeroIndex = 1
        while true do
            local rHeroActor = self:getActor("Hero"..nHeroIndex)
            if not rHeroActor then
                break
            else	
                table.insert(tHeroActors, rHeroActor)
                nHeroIndex = nHeroIndex + 1        
            end
        end
        return tHeroActors
    end

    function Ob:moveAllHeroesToSlot(tHeroActors, bSnap)
        for i, rHeroActor in ipairs(tHeroActors) do
            if rHeroActor:isMelee() then
                if bSnap then
                    self:snapActorToSlot("MeleeHeroInitLocs", rHeroActor)
                end
            else
                if bSnap then
                    self:snapActorToSlot("RangedHeroInitLocs", rHeroActor)
                end
            end
        end        
    end

    function Ob:moveAllEnemiesToSlot(tEnemies, bSnap)
        local nNumEnemies = table.getn(tEnemies)
        for i, rEnemy in ipairs(tEnemies) do
            if rEnemy:isMelee() then
                if bSnap then
                    self:snapActorToSlot("MeleeEnemyInitLocs", rEnemy)
                end
            else
                if bSnap then
                    self:snapActorToSlot("RangedEnemyInitLocs", rEnemy)
                end
            end
        end
    end

    function Ob:getTerritory()
        return self.rTerritory
    end

	function Ob:setSuccessful(bSuccessful)
		self.bSuccessful = bSuccessful 
	end

	function Ob:isSuccessful()
		return self.bSuccessful
	end

    function Ob:isCompleting()
        return self.bIsCompleting and not self.bHasCompleted
    end

    function Ob:hasCompleted()
        return self.bHasCompleted
    end

    function Ob:setCompleted(bSet)
        self.bHasCompleted = bSet
    end

    function Ob:setIsMissionEvent(bIsMissionEvent)
        self.tSaved.bIsMissionEvent = bIsMissionEvent
    end

    function Ob:isMissionEvent()
        return self.tSaved.bIsMissionEvent
    end

    function Ob:setTerritoryCleared(bCleared)
        self.bClearedTerritory = bCleared
    end

    function Ob:hasClearedTerritory()
        return self.bClearedTerritory
    end

    function Ob:setDifficultyLevel(eLevel)
        self.eRelDifficultyLevel = eLevel
    end

    function Ob:getDifficultyLevel()
        return self.eRelDifficultyLevel
    end

    function Ob:getCrimeLevelReduction()
        return self.rDef.nCrimeLevelReduction or 1
    end

	local kTEAMUP_BONUS = 5

	function Ob:createActors()
		-- create the hero
		self.tActors = {}
        local eMinMoraleToSet = nil
        for i, rHero in ipairs(self.tHeroes) do
            local rHeroActor = self:createEventActorFromHero(rHero)
            rHeroActor:setCurHP(rHero:getCurHP())
            self.tActors['Hero'..i] = rHeroActor
            if rHero:shouldImproveTeammateMorale() then
                eMinMoraleToSet = rHero:getMorale()
            end
            if rHeroActor:heroWasActive() then
                rHero:setActive(false)
            end
        end
        if eMinMoraleToSet then
            for i, rHero in ipairs(self.tHeroes) do
                rHero:setMorale(eMinMoraleToSet)
            end
        end

        local activeHeroes = g_Game:getActiveHeroes(true)
        
        for i, rHero in pairs(activeHeroes) do
            rHero:setSpriteActive(false)
        end
        
		if self.rDef and self.rDef.tActors then
            self.tEnemyActors = {}
            if self.rDef.tActors.Enemies then                
                for name, rActorDef in pairs(self.rDef.tActors.Enemies) do
                    self.tActors[name] = self:createEventActorFromDefKey(rActorDef.sCharDefKey)
                    table.insert(self.tEnemyActors, self.tActors[name])
                end
            end
            if self.rDef.tActors.OtherActors then                
               for name, rActorDef in pairs(self.rDef.tActors.OtherActors) do
                    self.tActors[name] = self:createEventActorFromDefKey(rActorDef.sCharDefKey)
                end
            end
		end
        
        
		for i, rActor in pairs(self.tActors) do
			rActor:setParent(self.rTransform)
            rActor:onEventStart()
		end
        
        -- TODO: pass in event type and actors if we want to get that granular
        g_Audio:onEnterEvent()
        
        --DFGraphics.dumpTextures("event_createActors")
        --g_Audio:dumpMemoryInfo("event_createActors")
	end

    function Ob:addActor(sCharDefKey, sName)
        if self:hasCompleted() then
            return nil
        end
        if self.tActors and sCharDefKey and sName then
            local nNumActors = util.countitems(self.tActors)
            local rNewActor = self:createEventActorFromDefKey(sCharDefKey)            
            if rNewActor then
                if self.tActors[sName] then
                    self.tActors[sName..tostring(nNumActors + 1)] = rNewActor
                else
                    self.tActors[sName] = rNewActor
                end
                if self.rTransform then
                    rNewActor:setParent(self.rTransform)
                end
                rNewActor:onEventStart()
                return rNewActor
            end
        end
        return nil
    end

    function Ob:addEnemy(sCharDefKey, sName)
        local rNewEnemy = self:addActor(sCharDefKey, sName)
        if rNewEnemy then
            if self.tFightEnemies then
                table.insert(self.tFightEnemies, rNewEnemy)
            end
        end
        return rNewEnemy
    end

	function Ob:cleanupActors()
		for i, rActor in pairs(self.tActors) do		
            rActor:removeFromLayer(self.rLayer)
            local rHero = rActor:getHero()
            if rHero then
                if rActor:heroWasActive() then
                    rHero:setActive(true, true)
                    rHero:runActivities(true)
                end
            end
            rActor:onEventEnd()
		end
		self.tActors = nil
        
        local activeHeroes = g_Game:getActiveHeroes(true, true)
        
        for i, rHero in pairs(activeHeroes) do
            rHero:setSpriteActive(true)
        end
        
        -- TODO: pass in event type and actors if we want to get that granular
        g_Audio:onEnterOffice()
        
        --DFGraphics.dumpTextures("event_cleanupActors")
        --g_Audio:dumpMemoryInfo("event_cleanupActors")
	end

    function Ob:getEventRenderLayer()
        return self.rLayer
    end

	function Ob:start(rLayer)
        MOAISim:forceGarbageCollection()
        Renderer.ShowBase(false)
        g_Audio:stopMusic()
        if self.rDef.sMusicCue then
            g_Audio:playMusic(self.rDef.sMusicCue, true)
        end
		self.rLayer = rLayer
        self.rBGLayer = Renderer.getLayer("EventBG").RenderLayer
		-- add the background
		self.rTransform = MOAITransform.new()

        self.bFleeRequested = false
        self.tQueuedHeroes = {}
        self.bAllowAddHeroes = true

		self:createActors()
        self:setUpBG()

        if not Event.staticTextures then
            Event.staticTextures = {}
            for i=1,3 do
                Event.staticTextures[i] = {}
                Event.staticTextures[i].tex = DFGraphics.loadTexture('MiscNonTPS/static'..((i == 1 and "") or tostring(i))..'.png')
                Event.staticTextures[i].tex:setWrap(true)
            end
        end


        g_MiddleManager:grantRenewableItems()
        
        local sAmbienceCueName = self.rDef.sAmbienceOverride or self.rEventBG:getAmbienceCue() 
        if sAmbienceCueName then
            g_Audio:playAmbience(sAmbienceCueName)
        end
        self.rEventUI = g_GUI:ShowEventUI(self)

        -- preload the fx sprite sheet
        self.sEffectSpritePath = "FX/FX"..Renderer.getTextureSizeSuffix()
        self.rEffectSprites = DFGraphics.loadSpriteSheet(self.sEffectSpritePath)

--        MOAISim:forceGarbageCollection()
        
        -- we use stats for robberies because they happen so darn much
        local sEventClassName = rDef.eventClassName
        if sEventClassName == "RobberyEvent" then
            DFAnalytics.incrStat("erob_strt", 1)
        else
            DFAnalytics.logEvent('evt_strt_' .. sEventClassName, { flavor = self.sDefKey } )
        end
        
        if table.getn(self.tHeroes) >= 4 then
            DFAchievement.updateAchievement('FourHeroesInFight', 1)
        end
        
        self.tSaved.bDecreaseReward = false
        self.nTotKORewards = 0
        
        local sMeteoriteKey, nLevel = Item.getRandomMeteoriteSpawn()
        if sMeteoriteKey and nLevel then
            self.rMeteoriteInfo = { sKey = sMeteoriteKey, nLevel = nLevel, bTapped = false }
        end

		self.rThread = MOAIThread.new()
		--self.rThread:run(self.eventThread, self)
		self.rThread:run(self.staticThread, self)
	end

    function Ob:addMeteorite()
        if self.rEventBG then
            if self.rMeteoriteInfo and self.rMeteoriteInfo.sKey and self.rMeteoriteInfo.nLevel then
                local tPos = self.rEventBG:getMeteoritePos()
                self.rMeteoriteInfo.rMeteor = Meteorite:new(self.rMeteoriteInfo)
                self.rMeteoriteInfo.rMeteor:setParent(self.rTransform)
                self.rMeteoriteInfo.rMeteor:insertIntoLayer(tPos)
            end        
        end
    end

    function Ob:staticThread()
        self.rEventUI:enableInteractiveUIElements(false)

        local rDeck = MOAIGfxQuad2D.new() 
        rDeck:setTexture(Event.staticTextures[1].tex)
        rDeck:setRect(0, -Renderer.getUIHeight(), Renderer.getUIWidth(), 0)

        rDeck:setUVRect(0,0,Renderer.getUIWidth() / 256, Renderer.getUIHeight() / 256)
        local rStatic = MOAIProp2D.new()
        rStatic:setDeck(rDeck)
        self.rEventUI:addElement(rStatic)
        rStatic:setLoc(0,0)
        rStatic:setPriority(-1)
        self.rEventUI:setElementsVisible(true)

        local i = 1
        local startTime = MOAISim:getElapsedTime()
        while MOAISim:getElapsedTime() - startTime < .3 do
            coroutine.yield()
            i = i+1
            if i > 3 then i = 1 end
            rDeck:setTexture(Event.staticTextures[i].tex)
        end
        self.rEventUI:removeElement(rStatic)
        
        Event.staticTextures = nil
        for i=1,3 do
            DFGraphics.unloadTexture('MiscNonTPS/static'..((i == 1 and "") or tostring(i))..'.png')
        end
        
        if self.rDef.eventClassName == "RobberyEvent" then
            self:addMeteorite()
        end
        self:eventThread()
    end

    function Ob:autoResolve()
        local bSuccessful = true
        self.bAutoResolve = true
        if self.customAutoResolveFncn then
            bSuccessful = self:customAutoResolveFncn()
        else
            bSuccessful = self:defaultAutoResolveFncn()
        end
        -- set the actor back active
        for i, rHeroActor in ipairs(self.tAutoResolveHeroes) do
            if rHeroActor:heroWasActive() then
                local rHero = rHeroActor:getHero()
                if rHero then
                    rHero:setActive(true)
                    rHero:runActivities(true)
                end
            end
        end
        self:setSuccessful(bSuccessful)
        self.bIsAutoResolving = false
        g_TodoList:onEventDelegated(self)
    end

    function Ob:canBeAutoResolved()
        if self.rDef.bNoAutoResolve then
            return false
        end
        return true
    end

    function Ob:queueToAutoResolve()
        self.bIsAutoResolving = true
        self.nTimeAtAutoResolve = MOAISim:getDeviceTime()
        -- we create the actors here
        -- create the actors
        self.tAutoResolveHeroes = {}
        for i, rHero in ipairs(self.tHeroes) do
            local rHeroActor = AutoResolveEventActor:new(self, rHero, nil, rHero.nPackLevel)
            rHeroActor:setCurHP(rHero:getCurHP())
            table.insert(self.tAutoResolveHeroes, rHeroActor)
            if rHeroActor:heroWasActive() then
                rHero:setActive(false)
            end
        end
        self.tAutoResolveEnemies = {}
		if self.rDef and self.rDef.tActors and self.rDef.tActors.Enemies then
            for name, rActorDef in pairs(self.rDef.tActors.Enemies) do
                local rDef = Character.getCharDef(rActorDef.sCharDefKey)
                if rDef then
                    local rEnemyActor = AutoResolveEventActor:new(self, nil, rDef, 1)
                    table.insert(self.tAutoResolveEnemies, rEnemyActor)
                end
            end
        end
    end

    function Ob:isAutoResolving()
        if self.bIsAutoResolving then
            local nResolveTime = Event.tMiscTweaks.kAUTO_RESOLVE_TIME or 60
            local nRemainingTime = nResolveTime - math.abs(MOAISim:getDeviceTime() - self.nTimeAtAutoResolve)
            if nRemainingTime <= 0 then
                nRemainingTime = 0
            end
            return true, nRemainingTime
        end
        return false, 0
    end

    function Ob:getOddsOfSuccess()
        -- REALLY simple guess
        if self.tHeroes then
            local tEnemyDefs = {}
            if self.rDef and self.rDef.tActors and self.rDef.tActors.Enemies then
                for name, rActorDef in pairs(self.rDef.tActors.Enemies) do
                    local rDef = Character.getCharDef(rActorDef.sCharDefKey)
                    if rDef then
                        table.insert(tEnemyDefs, rDef)
                    end
                end
            end
            local nTotalHeroHP = 0
            local nTotalHeroPower = 0
            for i, rHero in ipairs(self.tHeroes) do
                nTotalHeroHP = nTotalHeroHP + rHero:getCurHP()
                nTotalHeroPower = nTotalHeroPower + rHero:getStatPower()
            end            
            local nTotalEnemyHP = 0
            local nTotalEnemyPower = 0
            for i, rDef in ipairs(tEnemyDefs) do
                nTotalEnemyHP = nTotalEnemyHP + (rDef.tInitialStats[1].nMaxHP or 0)
                nTotalEnemyPower = nTotalEnemyPower + (rDef.tInitialStats[1].nPower or 0)
            end
            local nTotal = nTotalHeroHP + nTotalHeroPower + nTotalEnemyHP + nTotalEnemyPower
            if nTotal == 0 then
                nTotal = 10000 -- safety
            end
            return math.floor(((nTotalHeroHP + nTotalHeroPower) / nTotal) * 100)
        end
        return 100
    end

    function Ob:setUpBG()
        if self.rDef.bgDefFile then            
            self.rBackground = MOAIProp2D.new()
            self.rBackground:setDeck(g_GUI:GetOnePixelDeck())
            self.rBackground:setLoc(-Renderer.getUIWidth() / 2, Renderer.getUIHeight() / 2)
            self.rBackground:setScl(Renderer.getUIWidth(), Renderer.getUIHeight())
            self.rBackground:setColor(0, 0, 0)
            self.rBGLayer:insertProp(self.rBackground)

            local sPath = self.rDef.bgDefFile
            local tFields = DFUtil.split(sPath, ".")
            local sFilePath = tFields[1] or ''
            local sFileExtension = tFields[2] or 'lua'
            sPath = DFFile.getLevelPath(sFilePath..'_Med'..'.'..sFileExtension)
            if not sPath then
                return
            end
            local rBGDef = dofile(sPath)
            if not rBGDef then
                return
            end
            self.rEventBG = EventBG:new(rBGDef, self.rBGLayer)
            if not self.rEventBG then
                return
            end

            self.rEventBG:setLoc(0, 0)
            self.rEventBG:setParent(self.rTransform)
            self.rEventBG:setVisible(true)

            if self.rDef.tBGProps then
                for i, rPropInfo in ipairs(self.rDef.tBGProps) do
                    self.rEventBG:addBGProp(rPropInfo.sSpriteSheetPath, rPropInfo.sSpriteName, rPropInfo.x, rPropInfo.y)
                end
            end
        end
    end

    function Ob:getEventBG()
        return self.rEventBG
    end

    function Ob:clearBG()
        self.rEventBG:setVisible(false)
        self.rEventBG = nil
        self.rBGLayer:removeProp(self.rBackground)
    end

    function Ob:enableInteractiveUIElements(bEnable)
        self.rEventUI:enableInteractiveUIElements(bEnable)
    end

    function Ob:snapActorToSlot(slotName, rActor)
        if self.rEventBG then
            local rSlot = self.rEventBG:reserveNextSlot(slotName, rActor)
            if rSlot then
                local x, y = rSlot:getLoc()
                rActor:setLoc(x, y)
            end
        end
    end

    function Ob:moveActorToSlot(slotName, rActor, nDuration, sWalkAnim, bBlocking)
        if self.rEventBG then
            local rSlot = self.rEventBG:reserveNextSlot(slotName, rActor)
            if rSlot then
                local x, y = rSlot:getLoc()
                rActor:moveToLoc(x, y, nDuration, sWalkAnim, bBlocking)
            end
        end
    end

    function Ob:shakeThread(nMagnitude)
        local kIncrement = nMagnitude / 5
        local nDuration = 0.05
        while nMagnitude > 0 do
            self.rTransform:moveLoc(0, -nMagnitude, 0, nDuration, MOAIEaseType.LINEAR)
            util.sleep(nDuration)
            self.rTransform:moveLoc(0, nMagnitude, 0, nDuration, MOAIEaseType.LINEAR)
            util.sleep(nDuration)
            nMagnitude = nMagnitude - kIncrement
        end

        self.rTransform:setLoc(0, 0)

        self.rShakeThread = nil
    end

    function Ob:performScreenShake(nMagnitude)
        if not self.rShakeThread then
            self.rShakeThread = MOAIThread.new()
            self.rShakeThread:run(self.shakeThread, self, nMagnitude)
        end   
    end

    function Ob:abortEvent()
        for i, rHero in ipairs(self.tHeroes) do
            rHero:onAssignedToEvent(nil)
        end
		self.rThread = nil
		self.rTransform = nil        
        if self.rTerritory then
            self.rTerritory:onEventAborted(self)
        end
    end

	function Ob:onEnd()
        Renderer.ShowBase(true)

        self.bIsCompleting = true
        local tHeroActors = self:getAllHeroActors()
        for i, rHeroActor in ipairs(tHeroActors) do
            local nHP = math.floor(rHeroActor:getCurHP())
            if nHP <= 0 then
                nHP = 1
            end
            rHeroActor:getHero():setCurHP(nHP)
        end

        for i, rHero in ipairs(self.tHeroes) do
            rHero:onAssignedToEvent(nil)
            rHero:setPlayEnterAnimForNothingActivity(true)
        end

		self:cleanupActors()

        g_MiddleManager:clearRenewableItems()
        
        self.rEventUI = nil
		self.rThread = nil
		self.rTransform = nil

        self:clearBG()

        -- unload the preloaded spritesheet
        DFGraphics.unloadSpriteSheet(self.sEffectSpritePath)

        g_Audio:stopCurAmbience()
        g_Audio:stopMusic()

        MOAISim:forceGarbageCollection()

        self:setCompleted(true)

        if self.rMeteoriteInfo then
            if self.rMeteoriteInfo.bTapped and self.rMeteoriteInfo.sKey and self.rMeteoriteInfo.nLevel then
                g_MiddleManager:addMeteoriteToInventory(self.rMeteoriteInfo.sKey, self.rMeteoriteInfo.nLevel)
            end
            if self.rMeteoriteInfo.rMeteor then
                self.rMeteoriteInfo.rMeteor:removeFromLayer()
            end
        end

        g_CurrentMap:centerOnRoom('hub', 0) -- snap to the room       
	end

    function Ob:getMeteoriteInfo()
        return self.rMeteoriteInfo
    end

	function Ob:getActor(actorName)
		return self.tActors[actorName]
	end

    function Ob:attemptFlee(tHeroes)
        if self.bFleeRequested then
            -- fly out
            if tHeroes then
                for i, rHero in ipairs(tHeroes) do
                    rHero:exit()
                end
                util.sleep(3)
            end
            self.sWinner = "Enemy"
            return true
        end
        return false
    end

    function Ob:addQueuedHeroes(tHeroes)
        -- add extra
        local nIndex = 1
        while true do
            if nIndex > table.getn(self.tQueuedHeroes) then
                break
            end
            if not self.tQueuedHeroes[nIndex]:isTeleporting() then -- done teleporting
                self.tQueuedHeroes[nIndex]:attackNextTarget()
                table.insert(tHeroes, self.tQueuedHeroes[nIndex])
                table.remove(self.tQueuedHeroes, nIndex)
            else
                nIndex = nIndex + 1
            end
        end        
    end

    function Ob:getBestTarget(rAttacker)
        local tTargets = nil
        if rAttacker:isHero() then
            tTargets = self.tFightEnemies
        else
            tTargets = self.tFightHeroes
        end
        -- get the melee target first
        local rBestTarget = nil        
        if tTargets then
            local nMinNumAttackers = 10000
            for i, rTarget in ipairs(tTargets) do
                if rTarget:isMelee() then
                    local nNumAttackers = rTarget:getNumAttackers()
                    if nNumAttackers < nMinNumAttackers then
                        nMinNumAttackers = nNumAttackers
                        rBestTarget = rTarget
                    end
                end
            end
            -- on occasion, a ranged attacker will choose someone other than a melee target.
            if rBestTarget and not rAttacker:isHero() and not rAttacker:isMelee() then
                if math.random(1,2) == 2 then
                    rBestTarget = tTargets[math.random(1,#tTargets)]
                end
                --if math.random(1,3) == 3 then
                    --rBestTarget = nil
                --end
            end
            if rBestTarget then
                return rBestTarget
            else
                -- first ranged target
                return tTargets[1]
            end
        end        
    end

    function Ob:onKORewardAwarded(nAmount)
        if nAmount then
            self.nTotKORewards = self.nTotKORewards + nAmount
        end
    end

    function Ob:getTotKOReward()
        return self.nTotKORewards
    end

    function Ob:onActorKnockedOut(rActor)
        if self.actorKOCallback then
            self.actorKOCallback(self, rActor)
        end
        if rActor:isHero() then
            for i, rHero in ipairs(self.tFightHeroes) do
                if rHero == rActor then
                    table.remove(self.tFightHeroes, i)
                    if table.getn(self.tFightHeroes) == 0 then
                        -- player lost!
                        self.sWinner = "Enemy"
                    end
                    return
                end
            end
        else        
            for i, rEnemy in ipairs(self.tFightEnemies) do
                if rEnemy == rActor then
                    table.remove(self.tFightEnemies, i)
                    if table.getn(self.tFightEnemies) == 0 then
                        if table.getn(self.tEnemyWaves) == 0 then
                            -- player won!
                            self.sWinner = "Player"
                        else
                            self.tFightEnemies = table.remove(tWaves, 1)
                        end
                    end
                    return
                end
            end
        end
    end

    function Ob:getAllActiveHeroes()
        local tChars = {}
        for i, rHero in ipairs(self.tFightHeroes) do
            table.insert(tChars, rHero)
        end
        return tChars
    end

    function Ob:setHeroesBlocking(bBlocking)
        for i, rHero in ipairs(self.tFightHeroes) do
            rHero:setBlocking(bBlocking)
        end
    end

    function Ob:onHeroRoundComplete()
        self.nHeroRoundsComplete = self.nHeroRoundsComplete + 1
    end

    function Ob:onTurnElapsed()
        self.nTurnsSinceAbilityTrigger = self.nTurnsSinceAbilityTrigger + 1
        self.nRoundsComplete = self.nRoundsComplete + 1
    end

    function Ob:canTriggerAIAbility()
        if self.nTurnsSinceAbilityTrigger >= kDEFAULT_AITRIGGER_DELAY then
            return true
        end
        return false
    end

    function Ob:onAIAbilityTriggered()
        self.nTurnsSinceAbilityTrigger = 0
    end

    function Ob:getAllEnemies()
        local tChars = {}
        for i, rEnemy in ipairs(self.tFightEnemies) do
            table.insert(tChars, rEnemy)
        end
        return tChars
    end

    function Ob:playDefaultAbilityCue(sCueName)
        -- disabled due to untz handling of sound
        g_Audio:playCue(sCueName)
    end

    function Ob:startCrunchTime(nTimes)
        --self.bCrunchTime = true
        self.nCrunchTimes = nTimes
    end

    function Ob:setCombatPaused(bPaused)
        self.bCombatPaused = bPaused
    end

    function Ob:isPaused()
        return g_Game:getDebugPaused() or self.bCombatPaused
    end

    function Ob:heroesAlive()
        for _,rH in ipairs(self.tFightHeroes) do
            if not rH:isDead() then return true end
        end
        return false
    end

    function Ob:enemiesAlive()
        for _, rE in ipairs(self.tFightEnemies) do
            if not rE:isDead() then return true end
        end
        return false
    end

    function Ob:fightThread(tHeroes, tWaves, tCitizens, interruptionCB, bManualUIControl)
        self.sWinner = nil
        self.tFightHeroes = tHeroes

        if not bManualUIControl then self.rEventUI:enableInteractiveUIElements(true) end

        -- hack: I'd love to put this in EventUI:enableInteractiveUIElements, but tFightHeroes isn't
        -- always initialized by then.
        for _,v in ipairs(tHeroes) do
	        if v:getCurHP() < v:getMaxHP() then
                v:setHealthBarVisible(true)
                v.rHealthBar:update(v:getCurHP())
            end
        end

        self.tEnemyWaves = tWaves
        self.tFightEnemies = table.remove(tWaves, 1)
        self.nTurnsSinceAbilityTrigger = 0
        self.nHeroRoundsComplete = 0
        self.nRoundsComplete = 0
        local bEnemiesAlive = true

        while true do

            while self:isPaused() do
                coroutine.yield()
            end

            local bGoAnotherRound = true
            if self.nCrunchTimes and self.nCrunchTimes > 0 then
                self.nCrunchTimes = self.nCrunchTimes - 1
                self.bCrunchTime = true
            end

            if not self:heroesAlive() or not self:enemiesAlive() then
                break
            end

            while bGoAnotherRound do
                for i, rHero in ipairs(self.tFightHeroes) do        
                    if not rHero:isDead() and rHero:canAttack() then
                        if rHero:hasAnimation('attack_breathe') then
                            rHero:playAnimation('attack_breathe', true)
                        elseif rHero:hasAnimation('fight_breathe') then
                            rHero:playAnimation('fight_breathe', true)
                        end
                    end
                end

                for i, rHero in ipairs(self.tFightHeroes) do
                    if not rHero:isDead() then
                        util.sleep(0.2)
                        rHero:attackNextTarget()
                    end
                end
                local bHeroesAttacking = true
                while bHeroesAttacking do
                    if table.getn(self.tFightHeroes) <= 0 then
                        bHeroesAttacking = false
                    end
                    for i, rHero in ipairs(self.tFightHeroes) do
                        bHeroesAttacking = false
                        if rHero:isAttacking() then
                            bHeroesAttacking = true      
                            break
                        end
                    end
                    coroutine.yield()
                end
                while self:isPaused() do
                    coroutine.yield()
                end

                if not self:heroesAlive() or not self:enemiesAlive() then
                    break
                end

                bGoAnotherRound = false
                if self.bCrunchTime then
                    self.bCrunchTime = false
                    bGoAnotherRound = true
                end

                -- clear out the enemy anims
                for i, rEnemy in ipairs(self.tFightEnemies) do
                    if rEnemy:hasAnimation('fight_breathe') and not rEnemy:isDead() and rEnemy:canAttack() then
                        rEnemy:playAnimation('fight_breathe')
                    end
                end
            end
            if not self:heroesAlive() or not self:enemiesAlive() then
                break
            end
            self:onHeroRoundComplete()

            for i, rEnemy in ipairs(self.tFightEnemies) do
                if not rEnemy:isDead() then            
                    if not rEnemy:shouldManuallyTriggerAttack() then
                        util.sleep(0.2)
                        rEnemy:attackNextTarget()
                    end
                end
                while self:isPaused() do
                    coroutine.yield()
                end
            end
            if not self:heroesAlive() or not self:enemiesAlive() then
                break
            end

            local bEnemiesAttacking = true
            while bEnemiesAttacking do
                if table.getn(self.tFightEnemies) <= 0 then
                    bEnemiesAttacking = false
                end
                for i, rEnemy in ipairs(self.tFightEnemies) do
                    bEnemiesAttacking = false
                    if rEnemy:isAttacking() then
                        bEnemiesAttacking = true
                        break
                    end
                end
                coroutine.yield()
            end
            if not self:heroesAlive() or not self:enemiesAlive() then
                break
            end
            self:onTurnElapsed()

            while self:isPaused() do
                coroutine.yield()
            end

            if interruptionCB and interruptionCB(self) then
                self.sWinner = interruptionCB(self) 
                for i, rHero in ipairs(self.tFightHeroes) do
                    rHero:forceStopAttacking(true)
                end        
                for i, rEnemy in ipairs(self.tFightEnemies) do
                    if not rEnemy:shouldManuallyTriggerAttack() then
                        rEnemy:forceStopAttacking(true)
                    end
                end
                break
            end
        end

        if not self.sWinner then self.sWinner = (self:heroesAlive() and "Player") or "Enemy" end

        if tHeroes then
            for i, rHero in ipairs(tHeroes) do
                rHero:setHealthBarVisible(false)
            end
        end

        if self.tFightEnemies then
            for i, rEnemy in ipairs(self.tFightEnemies) do
                rEnemy:setHealthBarVisible(false)
            end
        end

        if not bManualUIControl then self.rEventUI:enableInteractiveUIElements(false) end

        local sEventClassName = self.rDef.eventClassName
        
        if self.sWinner == "Player" then    
            -- winning robberies is so frequent that it costs significant money to log it as telem events,
            --  roughly 10% of our total event logs, so we're switching it to a stat
            if sEventClassName == "RobberyEvent" then
                DFAnalytics.incrStat('erob_won', 1)
            else
                DFAnalytics.logEvent('evt_won_' .. sEventClassName, { flavor = self.sDefKey })
            end
            return true, tHeroes, nil, tCitizens
        else
            if self.bFleeRequested then
                DFAnalytics.logEvent('evt_fle_' .. sEventClassName, { flavor = self.sDefKey })
            else
                DFAnalytics.logEvent('evt_kia_' .. sEventClassName, { flavor = self.sDefKey })
            end
            
            return false, nil, self.tFightEnemies, tCitizens
        end                
    end

    function Ob:defaultAutoResolveFncn()
        local tHeroChars = DFUtil.deepCopy(self.tAutoResolveHeroes)
        local tEnemyChars = DFUtil.deepCopy(self.tAutoResolveEnemies)
        while true do
            if table.getn(tHeroChars) <= 0 then
                return false
            end
            if table.getn(tEnemyChars) <= 0 then
                return true
            end
            for i, rHero in ipairs(tHeroChars) do
                local bIsAlive, nTargetIndex = rHero:attackTarget(tEnemyChars, Event.tMiscTweaks.nAutoResolvePlayerDamageFactor, Event.tMiscTweaks.nAutoResolvePlayerAdditiveDamage)
                if not bIsAlive then
                    table.remove(tEnemyChars, nTargetIndex)
                end
            end
            for i, rEnemy in ipairs(tEnemyChars) do
                local bIsAlive, nTargetIndex = rEnemy:attackTarget(tHeroChars, Event.tMiscTweaks.nAutoResolveEnemyDamageFactor, Event.tMiscTweaks.nAutoResolveEnemyAdditiveDamage)
                if not bIsAlive then
                    table.remove(tHeroChars, nTargetIndex)
                end
            end
        end
    end

    -- blocking
    function Ob:typeSomeWords(words, nStartDelay, nLetterDelay)
        local sV = words
        self.rEventUI.rVictoryText:setString(" ")
        local boxx,boxy = self.rEventUI.rVictoryText:getLoc()
        local xmin,ymin,xmax,ymax = 0,0,0,0 -- self.rEventUI.rVictoryText:getStringBounds(1,1)

        local scl = self.rEventUI.rVictoryText:getScl()
        self.rEventUI.rCursor:setLoc(boxx+xmax,boxy)
        self.rEventUI.rStrangeRect:setLoc(boxx+xmin-33,boxy-57*scl)
        util.sleep(nStartDelay)
        local rTalkSound = g_Audio:playCue('MMOJ_Audio/UI/MissionDialogue_Lp', true)
        for i=1,string.len(sV) do
            self.rEventUI.rVictoryText:setString(string.sub(sV,1,i))
            xmin,ymin,xmax,ymax = self.rEventUI.rVictoryText:getStringBounds(1,i)
            -- bug in getStringBounds doesn't consider scale.
            xmin,ymin,xmax,ymax = xmin*scl,ymin*scl,xmax*scl,ymax*scl
            self.rEventUI.rCursor:setLoc(boxx+xmax,boxy)
            self.rEventUI.rStrangeRect:setLoc(boxx+xmin-33,boxy-57*scl)
            local cx,cy = xmax,(ymax-ymin)*.5
            util.sleep(nLetterDelay)
        end

        if rTalkSound then
            rTalkSound:stop()
        end
    end

    function Ob:playPlayerWonSequence(tHeroes, tCitizens, bHideVictoryText)
        -- won
        if tHeroes then
            for i, rHero in ipairs(tHeroes) do
                rHero:playAnimation("breathe", true, false)
            end
        end
        if tCitizens then
            for i, rCitizen in ipairs(tCitizens) do
                -- setup extra actors
                -- hack hack until we actually set these guys up as actors
                local bCitiFlipX = false
                if rCitizen.rDef == Character.getCharDef('CitiGirl') then
                    bCitiFlipX = true
                end                    
                rCitizen:playAnimation("cheer", bCitiFlipX, false)
            end
        end
        if tHeroes then
            for i, rHero in ipairs(tHeroes) do
                rHero:playAnimation("cheer", true, false)
                local sVictoryCue = rHero:getVictoryCue()
                if sVictoryCue then
                    g_Audio:playCue(sVictoryCue)
                end
                g_Audio:playCue(self.rDef.sKOSoundCue)
            end
        end

        if not bHideVictoryText then
            self:typeSomeWords("VICTORY!",1,.07)
        else
            util.sleep(.2)
        end
        util.sleep(1)
    end

    function Ob:playPlayerLostSequence(tEnemies, tCitizens, bNoPlayDialog)
        local tHeroes = self:getAllHeroActors()
        if tHeroes and not bNoPlayDialog then
            local sOutro = 'GenericLoseOutro'
            if tEnemies then
                local rRandomEnemy = tEnemies[math.random(table.getn(tEnemies))]
                if rRandomEnemy.rDef.sLoseOutro then
                    sOutro = rRandomEnemy.rDef.sLoseOutro
                else
                    local sCustomOutro = tHeroes[math.random(table.getn(tHeroes))].rDef.sLoseOutro
                    if sCustomOutro and math.random(1, 3) == 1 then
                        sOutro = sCustomOutro
                    end
                end
            end
            local rSpeakingEnemy = tEnemies[math.random(table.getn(tEnemies))]
            rSpeakingEnemy:sayLineFromDialogset(sOutro, 'upper', false, false)
        end
        -- lost
        if tEnemies then
            for i, rEnemy in ipairs(tEnemies) do
                if not rEnemy:isDead() then
                    if rEnemy:hasAnimation("cheer") then
                        rEnemy:playAnimation("cheer", false, false)
                        util.sleep(0.2)
                    end
                end
            end
        end
        if tCitizens then
            for i, rCitizen in ipairs(tCitizens) do
                -- setup extra actors
                -- hack hack until we actually set these guys up as actors
                local bCitiFlipX = true
                if rCitizen.rDef and rCitizen.rDef == Character.getCharDef('CitiMan') then
                    bCitiFlipX = false
                end       
                rCitizen:playAnimation("threathened_breathe", bCitiFlipX, false)
            end
        end
        self:typeSomeWords("DEFEAT!", .5, .05)
        util.sleep(1)
        g_GUI:HideDialogText()
        
        self.rEventUI.rVictoryText:setString("FEED TERMINATED")
        self.rEventUI.rCursor:setVisible(false)
        self.rEventUI.rStrangeRect:setVisible(false)

        -- failure bar #3 is full-screen, so make sure it doesn't cover things up.
        local nPri = self.rEventUI.rItemPaneBG:getPriority() or 0
        self.rEventUI.tFailureBars[3]:setPriority(nPri-1)

        for _,v in pairs(self.rEventUI.tFailureBars) do
            v:setVisible(true)
        end
        util.sleep(1)
        for _,v in pairs(self.rEventUI.tFailureBars) do
            v:setVisible(false)
        end
    end

    function Ob:createEventActorFromHero(rHero)
        local rActor = EventActor:new(self, rHero, nil, rHero.nPackLevel)
        
        return rActor
    end
    
    function Ob:createEventActorFromDefKey(sCharDefKey)
        local rDef = Character.getCharDef(sCharDefKey)
        if rDef then
            local rActor = EventActor:new(self, nil, rDef) 
            return rActor
        end
        return nil
    end

    function Ob:requestFlee(bRequest)
        self.bFleeRequested = bRequest
    end

    function Ob:setEndScreenCloseCB(fncn)
        self.endScreenCloseCB = fncn
    end

    function Ob:getEndScreenCloseCB()
        return self.endScreenCloseCB
    end

    function Ob:shouldShowNewspaper()
        if self.rDef.sWinHeadline or self.rDef.sLoseHeadline then
            return true
        end
        return false
    end

    function Ob:getNewspaperInfo()
        -- TEMP HACK
        if self:isSuccessful() then
            return self.sWinnerPortraitPath, self.sWinnerPortraitIndex, self.sLoserPortraitPath, self.sLoserPortraitIndex, self.rDef.sWinHeadline, self.rDef.sWinMainCaption, self.rDef.sWinSideCaption, self.rDef.nWinHeadlineFontSize, self.rDef.nWinHeadlineVerticalOffset
        else
            return self.sWinnerPortraitPath, self.sWinnerPortraitIndex, self.sLoserPortraitPath, self.sLoserPortraitIndex, self.rDef.sLoseHeadline, self.rDef.sLoseMainCaption, self.rDef.sLoseSideCaption, self.rDef.nLoseHeadlineFontSize, self.rDef.nLoseHeadlineVerticalOffset
        end
    end

    function Ob:onFinger(eventType, sx, sy)
        -- check to see if we've tapped something
        if self.rMeteoriteInfo and self.rMeteoriteInfo.rMeteor then
            local tLayer = Renderer.getLayer('EventFX')    
            local x, y = tLayer.RenderLayer:wndToWorld(sx, sy)
            local tProps = { tLayer.Partition:propListForPoint(x, y) }
            self.rMeteoriteInfo.rMeteor:onFinger(eventType, sx, sy, tProps)
        end
    end

	return Ob
end

return Event
