local Sprite = require('SpriteSheet')
local util = require("DFUtil")
local DFFile = require('DFCommon.File')
local Character = require("Character")
local Ability = require('Ability/Ability')
local Buff = require('Ability/Buff')
local Renderer = require('Renderer')
local EffectInstance = require('EffectInstance')
local ImageProgressBar = require('ui/ImageProgressBar')
local UIProgressBar = require('ui/UIProgressBar')
local DFGraphics = require('DFCommon.Graphics')
local DFAchievement = require('DFCommon.Achievement')
local DFAnalytics = require('DFCommon.Analytics')

local EventActor = {}

local kTARGET_XOFFSET = { -124, -120, -120, -128, -128, -118, -118, -130, -130, -134, -134  }
local kTARGET_YOFFSET = { 0, -40, 40, -55, 55, -70, 70, -85, 80, -90, 90 }
local kMOVEMENT_RATE = 200
local kMOVE_DIST_THRESH = 1

EventActor.STANCE_BALANCED = 1
EventActor.STANCE_OFFENSE = 2
EventActor.STANCE_DEFENSE = 3

function EventActor:new(rEvent, rHero, rDef, packLevel)
	local Ob = MOAITransform.new()

    Ob.__typename = "EventActor"
    
	function Ob:getName()
		return self.sName
	end

	function Ob:getIconDeck()
        return Character.getIconDeck(self.rDef)
	end

	function Ob:getStatLevel()
		return self.nLevel
	end

	function Ob:getStatPower()
		return (self.nPower + self.nPowerStatMod) * self.nStancePowerMod
	end

    function Ob:incrementPowerStatMod(nAmount)
        self.nPowerStatMod = self.nPowerStatMod + nAmount
    end

    function Ob:incrementCritStatMod(nAmount)
        self.nCritStatMod = self.nCritStatMod + nAmount
    end

    function Ob:incrementBlockMod(nAmount)
        self.nBlockMod = self.nBlockMod + nAmount
    end

	function Ob:setCurHP(nPts, bVisible)
        local nOld = self.nCurHP

		self.nCurHP = nPts

        if bVisible then
            self:setHealthBarVisible(true)
            self:setHealthText(self.nCurHP - nOld)
            self.rHealthBar:update(self.nCurHP)
        end
	end

	function Ob:setStatLevel(nLevel)
		self.nLevel = nLevel
	end
	
	function Ob:setStatPower(nPower)
		self.nPower = nPower
	end

    function Ob:getStatAwareness()
        return self.nAwareness
    end

    function Ob:getSpecialAbilityMult()
        return self.nSpecialAbilityMult
    end

    function Ob:isMelee()
        return self.rDef.bIsMelee
    end

    function Ob:isRanged()
        return false
    end
    
    function Ob:stunnable()
        return self.rDef.bStunnable
    end
    
    function Ob:getMovementRate()
        return kMOVEMENT_RATE
    end

    function Ob:cycleStance()
        if self.nCurStance == EventActor.STANCE_BALANCED then
            self.nCurStance = EventActor.STANCE_DEFENSE
            self.nStancePowerMod = .5
            self.nStanceDefenseMod = .6
        elseif self.nCurStance == EventActor.STANCE_DEFENSE then
            self.nCurStance = EventActor.STANCE_OFFENSE
            self.nStancePowerMod = 1.5
            self.nStanceDefenseMod = 1.6 
        elseif self.nCurStance == EventActor.STANCE_OFFENSE then
            self.nCurStance = EventActor.STANCE_BALANCED
            self.nStancePowerMod = 1
            self.nStanceDefenseMod = 1
        else
            assert(false)
        end
    end

    function Ob:getAttackCriticalMultiplier()
        -- for now random
        if self:isHero() then
            return self.rHero:getAttackCriticalMultiplier(self.nCritStatMod)
        else
            return 1.0
        end
    end

    function Ob:getNumAttackers()
        return self.tAttackers.n
    end

    function Ob:getMeleeAttackPos(rAttacker)
        if not self.tAttackers[rAttacker] then
            self.tAttackers[rAttacker] = 1
            self.tAttackers.n = self.tAttackers.n+1
        end

        local tXOffsetInfo = self.rDef.tXOffsetOverride or kTARGET_XOFFSET
        local tYOffsetInfo = self.rDef.tYOffsetOverride or kTARGET_YOFFSET
        local nNumAttackers = self.tAttackers.n
        local x, y = self:getLoc() 
        local xOffSet = tXOffsetInfo[nNumAttackers]
        if self:isHero() then
            xOffSet = 0 - xOffSet
        end
        local yOffset = tYOffsetInfo[nNumAttackers]
        return x + xOffSet, y + yOffset
    end

    function Ob:setBlocking(bBlocking)
        self.bIsBlocking = bBlocking
    end

    function Ob:isBlocking()
        return self.bIsBlocking
    end

    function Ob:attemptBlock()
        local nPct = self.nBlockChance + self.nBlockChanceMod
        if nPct then
            local nChance = (math.random(1, 10) / 10)
            if nChance <= nPct then
                return true
            end
        end
        return false
    end

    function Ob:canBlock()
        -- for now random
        if self:isHero() then
            return self:attemptBlock()
        else
            return false
        end
    end

    function Ob:addBlockEffectiveness(nAmount)
        self.nBlockEffectiveMod = self.nBlockEffectiveMod + nAmount
        if self.nBlockEffectiveMod < 0 then
            self.nBlockEffectiveMod = 0
        end
    end
    
    function Ob:addBlockChance(nAmount)
        self.nBlockChanceMod = self.nBlockChanceMod + nAmount
    end

    function Ob:isAttacking()
        if self.bIsAttacking and not self:isDead() then
            return true
        end
        return false
    end

    function Ob:triggerAbilityThread(rTarget, rAbility)
        if self.rEvent and rTarget and rAbility then
            rAbility:performAbility(self, rTarget, self.rEvent)
        end
    end

    function Ob:shouldManuallyTriggerAttack()        
        return self.rDef.bManuallyTriggerAttack
    end

    function Ob:onEventStart()
        for i, rAbility in pairs(self.tAbilities) do
            rAbility.rEventActor = self
            rAbility:resetNumTurns()
        end

    end

    function Ob:applyFightBuffs()
        if not self.bFightBuffsApplied then
            local tBuffsToApply = self:getFightBuffs()
            if tBuffsToApply then
                for i, rBuffInfo in ipairs(tBuffsToApply) do
                    if rBuffInfo.tBuffs then
                        for c, rBuff in ipairs(rBuffInfo.tBuffs) do
                            self:addBuff(rBuff, self)
                        end
                    end
                end
            end
            self.bFightBuffsApplied = true
        end
    end

    function Ob:onEventEnd()
        self.rFightThread = nil
        self.rRewardThread = nil
    
        for i, rAbility in pairs(self.tAbilities) do
            rAbility:resetNumTurns()
            rAbility.rEventActor = nil
        end
        if self.tBuffs then
            for i=#self.tBuffs,1,-1 do
            --for i, rBuff in ipairs(self.tBuffs) do  -- skips elements when buffs are removed.
                local rBuff = self.tBuffs[i]
                rBuff:endBuff()
            end
        end
    end

    function Ob:attackNextTarget()
        if not self.bFightBuffsApplied then
            self:applyFightBuffs()
        end
        self.bIsAttacking = true
        self.bForceStopped = false
        self.rFightThread = MOAIThread.new()
        self.rFightThread:run(self.attackNextTargetThread, self)
    end

    function Ob:updateTarget()
        if self.rTarget then
            if self.rTarget:isDead() then
                self.rTarget = nil
                self.bInRange = false
            end
        end
        if not self.rTarget then
            self.rTarget = self.rEvent:getBestTarget(self)
        end
    end

    function Ob:attackNextTargetThread()
        self:setHealthBarVisible(true)
        if self.rEvent then
            if not self:isDead() and not self.bForceStopped then                
                self:onTurnBegan()
                if not self:canAttack() then
                    util.sleep(0.25)
                else
                    self:updateTarget()
                    if not self.rTarget then
                        self.bIsAttacking = false
                        return
                    end
                    if self:isMelee() then
                        local targX, targY = self.rTarget:getMeleeAttackPos(self)
                        while targX and targY and not self.bInRange do
                            local x, y = self:getLoc()
                            local nDist = util.distance(x, y, targX, targY)
                            local nWalkDuration = nDist / kMOVEMENT_RATE
                            if nDist > kMOVE_DIST_THRESH then 
                                self:moveToMeleeTarget(self.rTarget, nWalkDuration, "run", true)
                                if self.rTarget:isDead() then
                                    self:updateTarget()
                                    if not self.rTarget then
                                        self.bIsAttacking = false
                                        return
                                    else
                                        targX, targY = self.rTarget:getMeleeAttackPos(self)
                                    end
                                else
                                    self.bInRange = true
                                end
                            else
                                self.bInRange = true
                            end
                        end
                    end
                    local rAbility = nil
                    local bAIAbilityTriggered = false
                    if self:isHero() then
                        rAbility = self:getQueuedAbility()
                        --assert(not self.rTarget or not self.rTarget:isDead())
                    else
                        if self.rEvent:canTriggerAIAbility() then
                            rAbility = self:getAIAbility()
                            bAIAbilityTriggered = true
                        end
                    end
                    if rAbility then
                        self.rQueuedAbility = nil
                    else
                        rAbility = self:getDefaultAbility()
                    end
                    if rAbility and rAbility:canPerformAbility() then
                        --assert(not self.rTarget or not self.rTarget:isDead())
                        rAbility:performAbility(self, self.rTarget, self.rEvent) -- this blocks
                    end
                    if bAIAbilityTriggered then
                        self.rEvent:onAIAbilityTriggered()
                    end
                    local bFlipX = false
                    if self:isHero() then
                        bFlipX = true
                    end
                    if self:hasAnimation('attack_breathe') then
                        self:playAnimation('attack_breathe', bFlipX)
                    elseif self:hasAnimation('fight_breathe') then
                        self:playAnimation('fight_breathe', bFlipX)
                    end
                end
                -- turn elapsed
                self:onTurnElapsed()
            end
        end
        self.bIsAttacking = false
    end

    function Ob:onTurnBegan()
        for i, rAbility in pairs(self.tAbilities) do
            if rAbility:isUnlocked() and rAbility.onTurnBegan then
                rAbility:onTurnBegan()
            end
        end
        --for i, rBuff in ipairs(self.tBuffs) do
        for i=#self.tBuffs,1,-1 do
            local rBuff = self.tBuffs[i]
            if rBuff.onTurnBegan then
                rBuff:onTurnBegan()
            end
        end
    end

    function Ob:onTurnElapsed()
        for i, rAbility in pairs(self.tAbilities) do
            if rAbility:isUnlocked() then
                rAbility:onTurnElapsed()
            end
        end
        --for i, rBuff in ipairs(self.tBuffs) do
        for i=#self.tBuffs,1,-1 do
            local rBuff = self.tBuffs[i]
            if rBuff then
                rBuff:onTurnElapsed()
            end
        end
    end

    function Ob:triggerAbility(sAbilityKeyName, rTarget)
        if sAbilityKeyName then
            if self.tAbilities[sAbilityKeyName] then
                if not rTarget then
                    rTarget = self.rEvent:getBestTarget(self)
                end
                if rTarget and not self:isDead() then
                    local rAbilityThread = MOAIThread.new()
                    rAbilityThread:run(self.triggerAbilityThread, self, rTarget, self.tAbilities[sAbilityKeyName])
                end
            end
        end
    end

    function Ob:isDead()
        return (self:getCurHP() <= 0)
    end    

	function Ob:getCurHP()
		return self.nCurHP
	end

    function Ob:addShield(nShieldLevel)
        self.nShieldLevel = self.nShieldLevel + nShieldLevel
        if self.nShieldLevel < 0 then
            self.nShieldLevel = 0
        end
    end   

    function Ob:addHealth(nHealthLevel, bVisible)
        if not self:isDead() then
            local nOld = self.nCurHP
            self.nCurHP = self.nCurHP + nHealthLevel
            if self.nCurHP > self.nMaxHP then
                self.nCurHP = self.nMaxHP
            end       
            self.rHealthBar:update(self.nCurHP)
            if bVisible then
                self:setHealthBarVisible(true)
                if self.nCurHP - nOld > 0 then
                    self:setHealthText(self.nCurHP - nOld)
                end
            end
        end
    end
    
    function Ob:addMana(nManaLevel, bVisible)
        if not self:isDead() then            
            local rHero = self:getHero()
            rHero:incrementMana(nManaLevel)
        end
    end

    function Ob:setHealthBarVisible(bVisible)
        if self.rHealthBar then
            self.rHealthBar:update(self.nCurHP)
            self.rHealthBar:setVisible(bVisible)
        end
    end

    function Ob:playHitEffect(rHitEffectDef)
        if rHitEffectDef and rHitEffectDef.sEffectPath then
            local rLayer = Renderer.getRenderLayer("EventFX")
            -- we assume one shots
            local rEffectInstance = EffectInstance:new(rHitEffectDef.sEffectPath, rLayer)
            if rEffectInstance then
                local xOffset = 0
                local yOffset = 0
                local nScale = 1
                if rHitEffectDef.offSetRange then
                    if rHitEffectDef.offSetRange.x then
                        xOffset = math.random(rHitEffectDef.offSetRange.x[1], rHitEffectDef.offSetRange.x[2])                        
                    end
                    if rHitEffectDef.offSetRange.y then
                        yOffset = math.random(rHitEffectDef.offSetRange.y[1], rHitEffectDef.offSetRange.y[2])                        
                    end
                end
                if rHitEffectDef.scaleRange then
                    -- since math.random requires integers
                    nScale = math.random(rHitEffectDef.scaleRange[1] * 100, rHitEffectDef.scaleRange[2]* 100)
                    nScale = nScale / 100
                end               
                rEffectInstance:setLoc(xOffset, yOffset)
                rEffectInstance:setScl(nScale, nScale)            
                rEffectInstance:setParent(self)
                rEffectInstance:start()
            end
        end
    end

    function Ob:setInvulnerable(bInvulnerable)
        self.bInvulnerable = bInvulnerable
    end

    function Ob:setImmortal(bImmortal)
        self.bImmortal = bImmortal
    end

    function Ob:getDefaultAttackMods()
        if self.rHero then
            return self.rHero:getDefaultAttackMods()
        end
        return nil
    end

    function Ob:getFightBuffs()
        if self.rHero then
            return self.rHero:getFightBuffs()
        end
        return nil
    end

    function Ob:incrementRevengeDamageAmount(nAmount)
        self.nRevengeDamageAmount = self.nRevengeDamageAmount + nAmount
        if self.nRevengeDamageAmount < 0 then
            self.nRevengeDamageAmount = 0
        end
    end

    function Ob:getVictoryCue()
        return self.rDef.sVictoryCue
    end

    function Ob:addVampirePct(nAmount)
        self.nVampirePct = self.nVampirePct + nAmount
        if self.nVampirePct > 1 then
            self.nVampirePct = 1
        end
        if self.nVampirePct < 0 then
            self.nVampirePct = 0
        end
    end

    function Ob:onDamageDealt(nAmount)
        if self.nVampirePct > 0 and nAmount then
            local nRefreshAmount = self.nVampirePct * nAmount
            if nRefreshAmount > 0 then
                self:addHealth(math.ceil(nRefreshAmount), true)
            end
        end
    end
    
	-- true: still alive
	-- false: dead
	function Ob:takeDamage(nPts, rHitEffectDef, rAttacker, bCritical)   
        if self.bInvulnerable then
            return true
        end
        if self:isDead() then
            return false
        end
        if rAttacker and self.nRevengeDamageAmount > 0 then
            self:attack(rAttacker, self.nRevengeDamageAmount)
        end
        if not self.rHealthBar:isVisible() then
            self.rHealthBar:setVisible(true)            
        end
        if self.bFled or self.bForceStopped then
            return true 
        end
        if self.nDodgeChance > 0 then
            local nChance = math.random(1, 100)
            if nChance <= (self.nDodgeChance * 100) then
                -- DODGED!!
                self:setOverlayText(g_DialogManager:getString("OVTX003TEXT"), 3, { 0, 1, 0 })
                return true
            end
        end

		-- spawn the damage
		local x, y = self:getLoc()
        nPts = nPts - self.nShieldLevel
        if nPts < 0 then
            nPts = 0
        end

        if self.rDef.rHitEffectDefOverride then
            self:playHitEffect(self.rDef.rHitEffectDefOverride)
        elseif rHitEffectDef then
            self:playHitEffect(rHitEffectDef)
        end
		
        -- play the knock out anim
        local bFlipX = false
        if self:isHero() then
            bFlipX = true
        end

		-- check for blocking attack
		self.bLastAttackWasBlocked = false
        if self:canBlock() then
            local nAmountToBlock = nPts * math.min(1, self.nBlockLevel + self.nBlockEffectiveMod)
            nPts = nPts - nAmountToBlock
            self:playAnimation('block', bFlipX, false)
            g_Audio:playCue('MMOJ_Audio/SFX/HeroSFX/Block_Melee')
			self.bLastAttackWasBlocked = true
        else
            if not self.rDef.bNoHitReact then
                self:playAnimation('hitreact', bFlipX, false)
            end
        end

        nPts = nPts * self.nStanceDefenseMod

		-- round (up)
		nPts = math.ceil(nPts)

		-- trigger critical sound if this is a crit strike
		if bCritical then
			g_Audio:playCue('MMOJ_Audio/SFX/HeroSFX/CriticalAttack_Stinger')
		end

        self:setHealthText(-nPts, bCritical)
		self.nCurHP = self.nCurHP - nPts
        if rAttacker and rAttacker.onDamageDealt then
            rAttacker:onDamageDealt(nPts)
        end
        self.rHealthBar:update(self.nCurHP)
		if self.nCurHP <= 0 then			
			self.nCurHP = 0
            if self.bImmortal and self.nCurHP <= 0 then
                self.nCurHP = 1
                return true
            end
            --for i, rBuff in ipairs(self.tBuffs) do
            for i=#self.tBuffs,1,-1 do
                local rBuff = self.tBuffs[i]
                rBuff:endBuff()
            end
            if self.rDef.tRewardOnKO then
                self.rKOingAttacker = rAttacker
                self:generateKOReward()                
            else
                self:playAnimation("knock_out", bFlipX, false)
            end
            if self.rDef.sCustomDeathThread then
                local threadFncn = EventActor.getCustomDeathThread(self.rDef.sCustomDeathThread)
                if threadFncn then
                    local rCustomThread = MOAIThread.new()
                    rCustomThread:run(threadFncn, self, self.rDef, self.rEvent)
                end                
            end
            self:setShadowVisible(false)
            if self.rDef.sKOSoundCue then
                g_Audio:playCue(self.rDef.sKOSoundCue)
            end
            if self.rEvent then
                self.rEvent:onActorKnockedOut(self)
            end
            if self.rDef.tKOAchievements then
                for i, sAchievementID in ipairs(self.rDef.tKOAchievements) do
                    DFAchievement.updateAchievement(sAchievementID, 1)
                end
            end
            if self.rHero then                
                self.rHero:setCurHP(1)
                self.rHero:decreaseMorale()
            end
            if self.rHealthBar:isVisible() then
                self.rHealthBar:setVisible(false)            
            end
			return false
		end
	end

    function Ob:rewardThread()
        local bFlipX = false
        if self:isHero() then
            bFlipX = true
        end
        self:playAnimation("knock_out", bFlipX, false, true)
        g_Audio:playCue('MMOJ_Audio/SFX/Battle/CoinDrop')
        local rEffectInstance = EffectInstance:new('Effects/Coins.lua', self.rRenderLayer)
        local x, y = self:getLoc()
        rEffectInstance:setLoc(x, y + 160)
        rEffectInstance:setParent(self)
        rEffectInstance:start()
        local rExplosionEffect = EffectInstance:new('Effects/Explosion.lua', self.rRenderLayer)
        local x, y = self:getLoc()
        rExplosionEffect:setLoc(x, y + 160)
        rExplosionEffect:setParent(self)
        rExplosionEffect:start()
        if self.rDef.tRewardOnKO then
            local nAmount = math.random(self.rDef.tRewardOnKO[1], self.rDef.tRewardOnKO[2])
            if nAmount > 0 then
                if self.rKOingAttacker then
                    local nKORewardPct = self.rKOingAttacker:getKORewardPct()
                    if nKORewardPct then
                        nAmount = math.ceil(nAmount * nKORewardPct)
                    end
                end
                g_Game:addCash(nAmount)
                if self.rEvent then
                    self.rEvent:onKORewardAwarded(nAmount)
                end
                DFAnalytics.incrStat('csh_ernd_kill', nAmount)
                local tColor = { 1, .91, .41 }
                if nAmount == 1 then
                    self:setOverlayText("+"..nAmount.." "..g_DialogManager:getString("OVTX002TEXT"), 3, tColor )
                else
                    self:setOverlayText("+"..nAmount.." "..g_DialogManager:getString("OVTX001TEXT"), 3, tColor )
                end
            end
        end
        util.sleep(0.3)
        self:setVisible(false)
    end

    function Ob:generateKOReward()
        self.rRewardThread = MOAIThread.new()
        self.rRewardThread:run(self.rewardThread, self)
    end

	function Ob:attack(rVictim, nDamageAmount, rHitEffectDef, bCritical)
		if rVictim then
			return rVictim:takeDamage(nDamageAmount, rHitEffectDef, self, bCritical)
		end
		return true
	end

	function Ob:getCurHP()
		return self.nCurHP
	end

	function Ob:getMaxHP()
		return self.nMaxHP
	end

    function Ob:setSpriteIndex(sIndexKey, flipX, flipY)
        for i, rSprite in ipairs(self.tSprites) do
            rSprite:setSpriteIndex(sIndexKey, flipX, flipY)
        end
    end

    function Ob:hasAnimation(animName)
        for i, rSprite in ipairs(self.tSprites) do
            if rSprite:getAnimation(animName) then
                return true
            end
        end
        return false
    end

    function Ob:playAnimation(animName, flipX, flipY, blocking)      
        if not self.bVisible then
            return 0
        end
        local nDuration = 0
        for i, rSprite in ipairs(self.tSprites) do
            if not rSprite:getAnimation(animName) then
                rSprite:setVisible(false)
            end
        end
        for i, rSprite in ipairs(self.tSprites) do
            if rSprite:getAnimation(animName) then
                rSprite:setVisible(true)                
                nDuration = rSprite:playAnimation(animName, flipX, flipY)
            end
        end
        if blocking and nDuration > 0 then
            util.sleep(nDuration)
        end
        return nDuration
    end

    function Ob:setVisible(bVisible)
        self.bVisible = bVisible
        for i, rSprite in ipairs(self.tSprites) do
            rSprite:setVisible(bVisible)
        end
    end

    function Ob:insertIntoLayer(rRenderLayer)
        if self.rRenderLayer ~= nil then
            self:removeFromLayer(self.rRenderLayer)
        end
        self.rRenderLayer = rRenderLayer
        if self.rShadowProp then
            self.rRenderLayer:insertProp(self.rShadowProp)
        end
        for i, rSprite in ipairs(self.tSprites) do
            self.rRenderLayer:insertProp(rSprite)
        end
        local rOverlayLayer = Renderer.getRenderLayer('EventFX')
        if rOverlayLayer then
            rOverlayLayer:insertProp(Ob.rCharOverlayText)
            rOverlayLayer:insertProp(Ob.rCharOverlayShadowText)  
            rOverlayLayer:insertProp(Ob.rHealthOverlayText)
            rOverlayLayer:insertProp(Ob.rHealthOverlayShadowText)
            if not self.rHealthBar then
                self.rHealthBar = ImageProgressBar:new(rOverlayLayer, self:getMaxHP(), "HealthBar_Green", "HealthBar_Red")
                self.rHealthBar:setParent(self)
                self.rHealthBar:setLoc(0, 160)
            end        
        end
    end
    
    function Ob:removeFromLayer(rRenderLayer)
        if self.rRenderLayer ~= rRenderLayer then
            print("WARNING: attempting to remove character from layer it's not in")
        end
        if not self.rRenderLayer then
            return
        end
        if self.rShadowProp then
            self.rRenderLayer:removeProp(self.rShadowProp)
        end
        for i, rSprite in ipairs(self.tSprites) do
            self.rRenderLayer:removeProp(rSprite)
            rSprite:unloadSpriteSheet()
        end
        local rOverlayLayer = Renderer.getRenderLayer('EventFX')
        if rOverlayLayer then
            rOverlayLayer:removeProp(Ob.rCharOverlayText)   
            rOverlayLayer:insertProp(Ob.rCharOverlayShadowText) 
            rOverlayLayer:removeProp(Ob.rHealthOverlayText)
            rOverlayLayer:removeProp(Ob.rHealthOverlayShadowText)
        end
        if self.rHealthBar then
            self.rHealthBar:setVisible(false)
        end
        if self.sCurPortraitPath then
            DFGraphics.unloadSpriteSheet(self.sCurPortraitPath)
        end
        self.sCurPortraitPath = nil
        self.rRenderLayer = nil
        for i, rAbility in pairs(self.tAbilities) do
            rAbility:unloadSounds()
        end
    end

    function Ob:setOverlayText(sText, nDuration, tColor)
        if sText then
            self.rCharOverlayText:setString(sText)
            self.rCharOverlayText:setColor(tColor[1], tColor[2], tColor[3])
            if nDuration > 0 then
                self.rCharOverlayText:seekColor(tColor[1], tColor[2], tColor[3], 0, nDuration)
            end

            self.rCharOverlayShadowText:setString(sText)
            self.rCharOverlayShadowText:setColor(0, 0, 0)
            if nDuration > 0 then
                self.rCharOverlayShadowText:seekColor(0, 0, 0, 0, nDuration)
            end
        end
    end

    function Ob:setHealthText(nAmount, bCritical)
        local sText = ""
        local nDuration = 3
        if bCritical and nAmount > 0 then
            sText = "CRIT!"

            self.rHealthOverlayShadowText:setColor(0, 0, 0)     
            self.rHealthOverlayShadowText:seekColor(0, 0, 0, 0, nDuration)
            self.rHealthOverlayShadowText:setString(sText)

            self.rHealthOverlayText:setColor(1, 1, .14)     
            self.rHealthOverlayText:seekColor(1, 0, .14, 0, nDuration)
            self.rHealthOverlayText:setString(sText)
        end        
    end

    function Ob:getHero()
        return self.rHero
    end

    function Ob:isHero()
        if self.rHero then
            return true
        else
            return false
        end
    end

    function Ob:heroWasActive()
        return self.bHeroWasActive
    end

    function Ob:enter(bBlocking)
        -- play the enter anim
        g_Audio:playCue('MMOJ_Audio/SFX/HeroSFX/LandIn')
        self:playAnimation("enter", true, false, bBlocking)
    end

    function Ob:exit(bBlocking)
        -- play the enter anim
        self.bFled = true
        self:playAnimation("exit", true, false, bBlocking)
    end

    function Ob:forceStopAttacking(bForceStopped)
        self.bForceStopped = bForceStopped
    end

    function Ob:flee(bBlocking)
        if self.rHero then
            self.rHero:improveMorale(kLOSS_FLEE_LOSS)
        end
        self:exit(bBlocking)
    end

    function Ob:moveToMeleeTarget(target, nDuration, sWalkAnim, bBlocking)
        local x, y = target:getMeleeAttackPos(self)
        local nCurX, nCurY = self:getLoc()
        local bFlipX = x > nCurX
        if sWalkAnim then
            self:playAnimation(sWalkAnim, bFlipX, false)
        end
        self:moveLoc(x - nCurX, y - nCurY, nil, nDuration, MOAIEaseType.LINEAR)
        if bBlocking then
			local nCurTime = MOAISim:getElapsedTime ()
            nDuration = nDuration + nCurTime
            while nCurTime < nDuration and not target:isDead() do
                coroutine.yield()
			    nCurTime = MOAISim:getElapsedTime ()
            end
            local newX,newY = self:getLoc()
            local targX, targY = target:getMeleeAttackPos(self)
            local blah = 1
            blah = blah+1
        end
    end

    function Ob:moveToLoc(x, y, nDuration, sWalkAnim, bBlocking)
        local nCurX, nCurY = self:getLoc()
        local bFlipX = x > nCurX
        if sWalkAnim then
            self:playAnimation(sWalkAnim, bFlipX, false)
        end
        self:moveLoc(x - nCurX, y - nCurY, nil, nDuration, MOAIEaseType.LINEAR)
        if bBlocking then
            util.sleep(nDuration)
        end
    end

    function Ob:teleportThread()
        -- fade in
        self:playAnimation("fight_breathe", true, false)
        self:setColor(1, 1, 1, 0)
        self:moveColor(1, 1, 1, 1, 1, true)
        self.rTeleportThread = nil
    end

    function Ob:setColor(r, g, b, a)
        for i, rSprite in ipairs(self.tSprites) do
            rSprite:setColor(r, g, b, a)
        end
    end

    function Ob:moveColor(r, g, b, a, nTime, bBlocking)
        for i, rSprite in ipairs(self.tSprites) do
            rSprite:moveColor(r, g, b, a, nTime)
        end
        if bBlocking then
            util.sleep(nTime)
        end
    end

    function Ob:isTeleporting()
        if self.rTeleportThread then
            return true
        else
            return false
        end            
    end

    function Ob:teleportIn(initX, initY, finalX, finalY, rRenderLayer)
        self:insertIntoLayer(rRenderLayer)
		self.rTeleportThread = MOAIThread.new()
		self.rTeleportThread:run(self.teleportThread, self, initX, initY, finalX, finalY)
    end

    function Ob:getDefaultAbility()
        return self.rDefaultAbility
    end

    function Ob:getActiveAbility()
        if self.rHero then
            return self.rHero:getActiveAbility()
        end
        return nil
    end

    function Ob:getAIAbility()
        local tPossibleAbilities = {}
        for i, rAbility in pairs(self.tAbilities) do
            if not rAbility:isDefault() then
                if not tPossibleAbilities then
                    tPossibleAbilities = {}
                end
                table.insert(tPossibleAbilities, rAbility)
            end
        end
        if not tPossibleAbilities then
            return nil
        end
        local nNumPossible = table.getn(tPossibleAbilities)
        if nNumPossible > 0 then
            return tPossibleAbilities[math.random(nNumPossible)]
        end
    end

    function Ob:getAbilities()
        return self.tAbilities
    end

    function Ob:getQueuedAbility()
        return self.rQueuedAbility
    end

    function Ob:setQueuedAbility(rAbility)
        self.rQueuedAbility = rAbility
    end

    function Ob:getBestAbility()
        for i, rAbility in pairs(self.tAbilities) do
            if not rAbility:isDefault() and rAbility:canPerformAbility() then
                return rAbility
            end
        end
        return self:getDefaultAbility()
    end

    function Ob:setCanAttack(bCanAttack)
        if bCanAttack then
            self.nNoAttackCounter = self.nNoAttackCounter - 1
            if self.nNoAttackCounter < 0 then
                self.nNoAttackCounter = 0
            end
        else
            self.nNoAttackCounter = self.nNoAttackCounter + 1
        end
    end

    function Ob:canAttack()
        return self.nNoAttackCounter == 0
    end

    function Ob:addBuff(rBuffDef, rSourceActor)
        if self:isDead() then
            return
        end
        if rBuffDef then
            local buffConstructor = Buff.getConstructor(rBuffDef.sBuffType)
            if buffConstructor then
                local rBuff = buffConstructor(nil, rBuffDef, self, rSourceActor or self)
                table.insert(self.tBuffs, rBuff)
                rBuff:applyBuff(self, self.rEvent)
            end
        end
    end

    function Ob:onBuffEnd(rBuff)
        for i, rCharBuff in ipairs(self.tBuffs) do
            if rBuff == rCharBuff then
                table.remove(self.tBuffs, i)
                return
            end
        end
    end

    function Ob:canPerformAbilities()
        if self:isDead() then
            return false
        end
--[[
        if self:isAttacking() then
            return true
        end
]]--
        return true
    end

    function Ob:sayLine(lineCode, sUpperOrLower, bBlocking, bWaitForInput)
        if lineCode then
            local sLine = g_DialogManager:getString(lineCode)
            if sUpperOrLower == 'upper' then
                g_GUI:DisplayUpperDialogText(self, sLine, bBlocking, bWaitForInput)
            else
                g_GUI:DisplayLowerDialogText(self, sLine, bBlocking, bWaitForInput)
            end            
        end
    end

    function Ob:sayLineFromDialogset(sSetName, sUpperOrLower, bBlocking, bWaitForInput)
        if not self.tDialogSets then
            self.tDialogSets = {}
        end
        if sSetName then
            local tDialogSet = self.tDialogSets[sSetName]
            if not tDialogSet then
                tDialogSet = g_DialogManager:getDialogSetInstance(sSetName)
                if not tDialogSet then
                    print("ERROR: Could not find the dialogset "..sSetName)
                    return
                end
            end

            local sLine = ""
            if tDialogSet.tLines then
                local nNumLines = table.getn(tDialogSet.tLines)
                if nNumLines > 0 then
                    if not tDialogSet.bPlaySequential then
                        local nIndex = math.random(nNumLines)
                        tDialogSet.nLastLineIndex = nIndex
                        sLine = g_DialogManager:getString(tDialogSet.tLines[nIndex].lineCode)
                    end
                end
            end
            if sUpperOrLower == 'upper' then
                g_GUI:DisplayUpperDialogText(self, sLine, bBlocking, bWaitForInput)
            else
                g_GUI:DisplayLowerDialogText(self, sLine, bBlocking, bWaitForInput)
            end
        end        
    end

    function Ob:getPortraitProp()
        if self.rPortraitSpriteSheet and self.sCostumeSet then
            local nIndex = self.rDef.tPortraitIndex[self.sCostumeSet]
            if nIndex then
                local rProp = MOAIProp2D.new()
                rProp:setDeck(self.rPortraitSpriteSheet)
                rProp:setIndex(self.rPortraitSpriteSheet.names[nIndex])
                return rProp
            end
        end
        return nil
    end

    function Ob:getCurPortraitPath()
        return self.rDef.tPortraitPaths[self.sCostumeSet], self.rDef.tPortraitIndex[self.sCostumeSet]
    end

    function Ob:setShadowVisible(bVisible)
        if self.rShadowProp then
            self.rShadowProp:setVisible(bVisible)
        end
    end

    function Ob:addKORewardPct(nPct)
        if nPct then
            self.nKORewardPct = self.nKORewardPct + nPct
        end
    end

    function Ob:getKORewardPct()
        return self.nKORewardPct
    end

    function Ob:addDodgeChance(nPct)
        self.nDodgeChance = self.nDodgeChance + nPct
    end

    Ob.tSprites = {}
    Ob.rHero = rHero
    Ob.rEvent = rEvent   
    Ob.nNoAttackCounter = 0
    Ob.tBuffs = {}
    Ob.nShieldLevel = 0
    Ob.bFled = false
    Ob.tUsedAbilities = {}
    Ob.bVisible = true
    Ob.bIsBlocking = false

    local sSpriteSheetKey = rEvent:getSpriteSheetKey()

    local tSpriteSheetDefPaths = nil
    local sAnimDefFileName = nil
    if rHero then
        rDef = rHero.tCharDef
        Ob.tAbilities = rHero:getAbilities()
        Ob.rDefaultAbility = rHero:getDefaultAbility()
        for i, rAbility in pairs(Ob.tAbilities) do
            rAbility.rCharacter = rHero
--            rAbility:preloadSounds()
        end
    else
        -- setup the ablities manually
        -- initialize the abilities
        Ob.tAbilities = {}
        Ob.rDefaultAbility = nil
        if rDef.tAbilityDefKeys then
            for i, sAbilityDefKey in ipairs(rDef.tAbilityDefKeys) do
                local rAbilityDef = Character.getAbilityDef(sAbilityDefKey)
                if rAbilityDef then
                    if rAbilityDef.sAbilityType then
                        local constructor = Ability.getConstructor(rAbilityDef.sAbilityType)
                        if constructor then
                            local rAbility = constructor(nil, rAbilityDef, Ob, nil, nil, sAbilityDefKey)
                            if rAbility:isDefault() then
                                Ob.rDefaultAbility = rAbility
                            end
                            Ob.tAbilities[sAbilityDefKey] = rAbility    
--                            rAbility:preloadSounds()
                        end
                    end
                end
            end
        end
    end
    Ob.rQueuedAbility = nil

    local costumeSetName = 'default'
    if rHero then
        costumeSetName = rHero:getCostumeSetName()
    end
    if sSpriteSheetKey and rDef[sSpriteSheetKey] then        
        tSpriteSheetDefPaths = rDef[sSpriteSheetKey].tSpriteSheetDefPaths[costumeSetName]
        sAnimDefFileName = rDef[sSpriteSheetKey].sAnimDefFileName
    end
    if not tSpriteSheetDefPaths then
        tSpriteSheetDefPaths = rDef.tSpriteSheetDefPaths[costumeSetName]
        sAnimDefFileName = rDef.sAnimDefFileName
    end
    Ob.sCostumeSet = costumeSetName

    if tSpriteSheetDefPaths then
		if not rDef.bNoShadow then
			Ob.rShadowProp = Character.getShadowProp()
			Ob.rShadowProp:setParent(Ob)
		end

        local nExtraScale = 1
        if rDef and rDef.tScaleInfo then
            if Renderer.getCurScreenInfo() and Renderer.getCurScreenInfo().key then
                if rDef.tScaleInfo[Renderer.getCurScreenInfo().key] then
                    nExtraScale = rDef.tScaleInfo[Renderer.getCurScreenInfo().key]
                end
            end
        end

        for i, sSpriteSheetDefPath in ipairs(tSpriteSheetDefPaths) do
            local rSprite = nil
            rSprite = Sprite:new(sSpriteSheetDefPath, sAnimDefFileName, rDef.centerToBaseX, rDef.centerToBaseY)    
            rSprite.rCharEnt = Ob
            rSprite:setParent(Ob)
            
            rSprite:setScl(nExtraScale)
            table.insert(Ob.tSprites, rSprite)                
        end

        if rDef.tAdditionalSpriteSheets then
            for i, rAddSpriteInfo in ipairs(rDef.tAdditionalSpriteSheets) do
                local rSprite = nil
                rSprite = Sprite:new(rAddSpriteInfo.sSpriteSheetDefPath, rAddSpriteInfo.sAnimDefFileName)                
                rSprite.rCharEnt = Ob
                rSprite:setParent(Ob)

                if rAddSpriteInfo.nYOffset then
                    rSprite:setLoc(0, rAddSpriteInfo.nYOffset)
                end
                
                rSprite:setScl(nExtraScale)
                table.insert(Ob.tSprites, rSprite)
            end
        end
    end

    if rDef.tPortraitPaths and rDef.tPortraitPaths[costumeSetName] then
        local sPath = rDef.tPortraitPaths[costumeSetName]..Renderer.getTextureSizeSuffix()
        Ob.rPortraitSpriteSheet = DFGraphics.loadSpriteSheet(sPath)
        if Ob.rPortraitSpriteSheet then
            for sName, data in pairs(Ob.rPortraitSpriteSheet.names) do
                DFGraphics.alignSprite(Ob.rPortraitSpriteSheet, sName, "left", "top", Renderer.getTextureSizeFactor(), Renderer.getTextureSizeFactor())
            end                    
            Ob.sCurPortraitPath = sPath
        end
    end

	Ob.sName = rDef.sName
	Ob.rDef = rDef

	packLevel = packLevel or 1

    Ob.nRevengeDamageAmount = 0
    Ob.tAttackers = {}
    Ob.tAttackers.n = 0
	-- stats
	Ob.nLevel = rDef.tInitialStats[packLevel].nLevel
	Ob.nPower = rDef.tInitialStats[packLevel].nPower
    Ob.nSpecialAbilityMult = 1

	Ob.nCurHP = rDef.tInitialStats[packLevel].nMaxHP * g_Game.GLOBAL_HEALTH_MOD
	Ob.nMaxHP = rDef.tInitialStats[packLevel].nMaxHP * g_Game.GLOBAL_HEALTH_MOD
    Ob.nAwareness = rDef.tInitialStats[packLevel].nAwareness
    Ob.nCurStance = EventActor.STANCE_BALANCED
    Ob.nStancePowerMod = 1
    Ob.nStanceDefenseMod = 1

    Ob.bHeroWasActive = false
    if rHero then
        Ob.nLevel = rHero:getStatLevel()
        Ob.nPower = rHero:getStatPower()
        Ob.nSpecialAbilityMult = rHero:getSpecialAbilityMult()
        Ob.nCurHP = rHero:getCurHP()
        Ob.nMaxHP = rHero:getMaxHP()
        Ob.nAwareness = rHero:getStatAwareness()
        Ob.nShieldLevel = rHero:getStatMod('nShield')
        Ob.nBlockLevel = rHero:getBlockEffectiveness()
        Ob.nBlockChance = rHero:getBlockChance()
        if rHero:isActive() then
            Ob.bHeroWasActive = true
        end
    end

    Ob.nPowerStatMod = 0
    Ob.nCritStatMod = 0
    Ob.nBlockEffectiveMod = 0
    Ob.nBlockChanceMod = 0
    Ob.nVampirePct = 0
    
    if Ob.rDef.sDefaultSpriteIndex then
        Ob:setSpriteIndex(Ob.rDef.sDefaultSpriteIndex)
    end

    if Ob.rDef.initalAnimation then
        local bFlipX = false
        if Ob:isHero() then
            bFlipX = true
        end
        Ob:playAnimation(Ob.rDef.initalAnimation, bFlipX, false)
    end

    Ob.rCharOverlayShadowText = g_GUI:CreateTextBox(35, MOAITextBox.CENTER_JUSTIFY, 'bangers')
    Ob.rCharOverlayShadowText:setRect(-300, 80, 300, 0)
	Ob.rCharOverlayShadowText:setLoc(2, 216)
    Ob.rCharOverlayShadowText:setColor(1, 0, 0)
	Ob.rCharOverlayShadowText:setParent(Ob)

    Ob.rCharOverlayText = g_GUI:CreateTextBox(35, MOAITextBox.CENTER_JUSTIFY, 'bangers')
    Ob.rCharOverlayText:setRect(-300, 80, 300, 0)
	Ob.rCharOverlayText:setLoc(0, 214)
    Ob.rCharOverlayText:setColor(1, 0, 0)
	Ob.rCharOverlayText:setParent(Ob)

    Ob.rHealthOverlayShadowText = g_GUI:CreateTextBox(35, MOAITextBox.CENTER_JUSTIFY, 'bangers')
    Ob.rHealthOverlayShadowText:setRect(-300, 80, 300, 0)
	Ob.rHealthOverlayShadowText:setLoc(72, 182)
    Ob.rHealthOverlayShadowText:setColor(1, 0, 0)
	Ob.rHealthOverlayShadowText:setParent(Ob)

    Ob.rHealthOverlayText = g_GUI:CreateTextBox(35, MOAITextBox.CENTER_JUSTIFY, 'bangers')
    Ob.rHealthOverlayText:setRect(-300, 80, 300, 0)
	Ob.rHealthOverlayText:setLoc(70, 180)
    Ob.rHealthOverlayText:setColor(1, 0, 0)
	Ob.rHealthOverlayText:setParent(Ob)

    Ob.bImmortal = false
    Ob.bIsBlocking = false
	Ob.bLastAttackWasBlocked = false
    Ob.nDodgeChance = 0

    Ob.nKORewardPct = 1

	return Ob
end

local function spawnCitizen(rEventActor, rDef, rEvent)    
    if rEventActor and rEvent and rDef and rDef.sCitizenCharDefToSpawn then
        local rCitizen = rEvent:addActor(rDef.sCitizenCharDefToSpawn, 'SpawnedCitizen')
        if rCitizen then
            local rLayer = rEvent:getEventRenderLayer()
            if rLayer then
                local x, y = rEventActor:getLoc()
                rCitizen:setLoc(x, y)
                rCitizen:insertIntoLayer(rLayer)
				local rEffectInstance = EffectInstance:new('Effects/AlienBodyDeath.lua', Renderer.getRenderLayer("EventFX"))
				rEffectInstance:setLoc(x, y+160)
				rEffectInstance:setScl(2, 2)  
				rEffectInstance:setParent(rCitizen)
				rEffectInstance:start()
                local bFlipX = false
                if rDef.sCitizenCharDefToSpawn == 'CitiGirl' then
                    bFlipX = true
                end
                rCitizen:playAnimation("cheer", bFlipX, false)
            end
        end
    end
end

local function spawnEnemy(rEventActor, rDef, rEvent)    
    if rEventActor and rEvent and rDef and rDef.sEnemyCharDefToSpawn then
        local rEnemy = rEvent:addEnemy(rDef.sEnemyCharDefToSpawn, 'SpawnedEnemy')
        if rEnemy then
            local rLayer = rEvent:getEventRenderLayer()
            if rLayer then
                local x, y = rEventActor:getLoc()
                rEnemy:setLoc(x, y)
                rEnemy:insertIntoLayer(rLayer)
            end
        end
    end
end

local function loveGunExplosion(rEventActor, rDef, rEvent)    
    if rEventActor and rEvent and rDef then
        local rLayer = rEvent:getEventRenderLayer()
        if rLayer then
            local x, y = rEventActor:getLoc()
            rEventActor:setLoc(x, y - 100)
            rEventActor:setScl(0.5)
            local rEffectInstance = EffectInstance:new('Effects/Explosion.lua', Renderer.getRenderLayer("EventFX"))
            rEffectInstance:setLoc(x, y - 50)
            rEffectInstance:setScl(3, 3)
            rEffectInstance:start()
        end
    end    
end


EventActor.tCustomDeathThread =
{
   spawnCitizen = spawnCitizen,
   spawnEnemy = spawnEnemy,
   loveGunExplosion = loveGunExplosion,
}

function EventActor.getCustomDeathThread(sName)
    return EventActor.tCustomDeathThread[sName]
end

return EventActor
