-- Singleton object for input handling
-- Instance vars:
--   focus      The library (or whatever) that consumed a DOWN event.
--              Only it will get the following MOVE and UP events.

local input = {}

local character = require("Character")
local Renderer = require('Renderer')
local MiddleManager = require('MiddleManager')
local Activity = require('activity/Activity')
local WalkToRoomTask = require('WalkToRoomTask')
local WalkToTileTask = require('WalkToTileTask')
local StartEffectTask = require('StartEffectTask')
local DFMath = require("DFCommon.Math")
local DFAnalytics = require('DFCommon.Analytics')

-- Just for laziness' sake
local TOUCH_DOWN = MOAITouchSensor.TOUCH_DOWN
local TOUCH_MOVE = MOAITouchSensor.TOUCH_MOVE
local TOUCH_UP = MOAITouchSensor.TOUCH_UP

-- Internal: mouse state, not used on iOS devices
local m_x = 0
local m_y = 0
local m_down = false

-- ----------------------------------------------------------------------
-- Input and focus handling
-- ----------------------------------------------------------------------

function input:init()
    -- the guy who wants MOVE/UP events
    self.focus = nil

    if MOAIInputMgr.device.pointer then
        -- called on button state _change_ (not continuously)
        local function mouseLeftCb(down)
            -- nb: self is an upvalue
            m_down = down
            local evt
            if down then evt = TOUCH_DOWN else evt = TOUCH_UP end
            self:onFinger(evt, m_x, m_y)
        end
		
        local function mouseRightCb(down)
            -- debug code
			--rDude:setPosition(x, y, true)
        end

        -- called on pointer position _change_ (not continuously)
        local function pointerCb(x,y)
            -- nb: self is an upvalue
            m_x, m_y = x, y
            if m_down then
                self:onFinger(TOUCH_MOVE, m_x, m_y)
            end
        end

        MOAIInputMgr.device.pointer:setCallback ( pointerCb )
        MOAIInputMgr.device.mouseLeft:setCallback ( mouseLeftCb )
		MOAIInputMgr.device.mouseRight:setCallback ( mouseRightCb )
    else
        -- TODO: ignore MOVE and UP events except for idx associated
        -- with the most-recent TOUCH_DOWN?
        local function touchCb(eventType, idx, x,y, tapCount)
            -- nb: self is an upvalue
            if (eventType == TOUCH_DOWN or
                eventType == TOUCH_MOVE or
                eventType == TOUCH_UP) then
                self:onFinger(eventType, x,y)
            end
        end
        MOAIInputMgr.device.touch:setCallback(touchCb)
    end

    -- KEYBOARD IS MAINLY FOR DEBUG
    local function keyboardCb(key, bDown)
        self:onKeyboard(key, bDown)
    end    

    if MOAIInputMgr.device.keyboard then
        MOAIInputMgr.device.keyboard:setCallback(keyboardCb)
    end
end

-- Tries hard to be a (simplified) copy of the touch callback API:
--   function(eventType, idx, x,y, tapCount)
--
-- eventType is:
--   TOUCH_DOWN         Transition from no-touch to touch
--   TOUCH_MOVE         Movement while touching
--   TOUCH_UP           Transition from touch to no-touch
-- 

g_TouchMoved = false
g_TouchStartX = -9999 -- arbitrary, invalid value
g_TouchStartY = -9999

-- this is tunable! If users complain about failed touches this is something we should tune.
kTOUCHMOVE_SCROLL_THRESHOLD = 28


function input:onFinger(eventType, sx, sy)

    -- need to divide the pixel movement by the screen scale (so devices like the iPad 3 don't scroll super fast)
    local nScreenScaleX, nScreenScaleY = Renderer.getDisplayScl()

	--[[
	if eventType == MOAITouchSensor.TOUCH_DOWN then	print("TOUCH_DOWN") end
	if eventType == MOAITouchSensor.TOUCH_MOVE then	print("TOUCH_MOVE") end
	if eventType == MOAITouchSensor.TOUCH_UP then print("TOUCH_UP") end
	]]--

	if eventType == MOAITouchSensor.TOUCH_DOWN then
        g_TouchMoved = false
        g_TouchStartX = sx
        g_TouchStartY = sy
    elseif eventType == MOAITouchSensor.TOUCH_MOVE then
        local touchStartDistDX = (sx - g_TouchStartX) / nScreenScaleX
        local touchStartDistDY = (sy - g_TouchStartY) / nScreenScaleY
    
        local touchStartDist = math.sqrt((touchStartDistDX * touchStartDistDX) + (touchStartDistDY * touchStartDistDY))
        
        -- add a slight deadzone so shaky fingers can still touch stuff
        if touchStartDist > kTOUCHMOVE_SCROLL_THRESHOLD then
            g_TouchMoved = true
        end
	end
	
	local handled = false

    if g_eventManager then
        g_eventManager:onFinger(eventType, sx, sy)
    end

    if g_GUI:onFinger(eventType, sx, sy) then    
        return
    end
	
    -- see if we have two touches
    if MOAIInputMgr.device.touch then
        if g_Game:isActionEnabled('CameraControl') then
            local nTouchIndex1, nTouchIndex2 = MOAIInputMgr.device.touch:getActiveTouches()
            if nTouchIndex1 and nTouchIndex2 then
                local x1, y1 = MOAIInputMgr.device.touch:getTouch(nTouchIndex1)
                local x2, y2 = MOAIInputMgr.device.touch:getTouch(nTouchIndex2)
                local nTouchDist = DFMath.distance(x1, y1, x2, y2)
                if not g_lastTouchDist then
                    g_lastTouchDist = nTouchDist
                end
                if g_lastTouchDist ~= nTouchDist then
                    -- zoom in or out
                    local x, y = Renderer:getCurZoomScl()
                    local nDiff = g_lastTouchDist - nTouchDist
                    local nIncrement = nDiff * 0.005
                    Renderer.setCameraScl(x + nIncrement, y + nIncrement, 0)                    
                end
                g_lastTouchDist = nTouchDist
                return
            else
                if g_lastTouchDist then
                    g_lastTouchDist = nil
                    return
                end
            end         
        end
    end
    
    if not g_TouchMoved then
        local numLayers = Renderer.getNumLayers()
        for i=1,numLayers do
            local index = numLayers - i + 1
            local tLayer = Renderer.getLayerByIndex(index)
            local x, y = tLayer.RenderLayer:wndToWorld(sx, sy)
            local tProps = { tLayer.Partition:propListForPoint (x, y) }
            for j,rSelection in ipairs(tProps) do
                local rSelection = tProps[j]
                if rSelection and rSelection.onFinger then
                    handled = rSelection:onFinger(eventType, sx, sy)
                end
            end
            if handled then
                break
            end
        end
        
        -- Get tile information
        if not handled and not g_TouchMoved and eventType == MOAITouchSensor.TOUCH_UP then                
                   
            handled = g_CurrentMap:onFinger(eventType, sx, sy)
            
            if not handled then
                local tLayer = Renderer.getLayer("Gameplay")
                local x, y = tLayer.RenderLayer:wndToWorld(sx, sy)
                local rRoom = g_CurrentMap:getRoomAtCoord(x, y)
                
                if rRoom then
                    handled = rRoom:onFinger(eventType, sx, sy)
                end           
            end
        end
    end
    
    -- Scroll the camera
    if not handled then
       
        if g_Game:isActionEnabled('CameraControl') then
            if eventType == MOAITouchSensor.TOUCH_DOWN then
                g_lastSx = sx
                g_lastSy = sy
		
            end
            if eventType == MOAITouchSensor.TOUCH_MOVE then
	    
                if not g_lastSx then g_lastSx = sx end
                if not g_lastSy then g_lastSy = sy end
                local dx = sx - g_lastSx
                local dy = sy - g_lastSy
                
                local dist = math.sqrt(dx * dx + dy *dy)
                
                if g_TouchMoved or dist > kTOUCHMOVE_SCROLL_THRESHOLD then
                    dx = dx / nScreenScaleX
                    dy = dy / nScreenScaleY
                
                    Renderer.moveCamera(-dx, dy, 0, true)
                    
                    g_TouchMoved = true
                    
                    g_lastSx = sx
                    g_lastSy = sy
                end
            end
        end
    end
    
    if eventType == MOAITouchSensor.TOUCH_UP then
        g_TouchMoved = false
    end
	
end

function input:onKeyboard(key, bDown)
    local nDuration = 0.1
    local nZoomIncrement = 0.05
    if bDown then
        if self.DBG_awaitingInput then
            if self.DBG_awaitingInput == string.byte("L") then
                if key >= string.byte("0") and key <= string.byte("9") then
                    local lvl = key - string.byte("1")+1
                    if lvl == 0 then lvl = 10 end
                    local tHeroes = g_Game:getActiveHeroes()
                    for _,v in ipairs(tHeroes) do
                        --v.tSaved.tStats.nLevel = lvl
                        v.tSaved.tStats.nLevel = 1
                        v.tSaved.tStats.nPower = v.tCharDef.tInitialStats[v.nPackLevel].nPower
                        v.tSaved.nNumTrainingSessions = 0
                        v.tSaved.tStats.nAwareness = v.tCharDef.tInitialStats[v.nPackLevel].nAwareness
                        v.tSaved.tStats.nMaxHP = v.tCharDef.tInitialStats[v.nPackLevel].nMaxHP
                        v.tSaved.tStats.nCharisma = v.tCharDef.tInitialStats[v.nPackLevel].nCharisma
                        if lvl > 1 then
			                v.tSaved.tStats.nXP = character.tLevelImprovementInfo[lvl - 1].nXPForNextLevel
                        else
                            v.tSaved.tStats.nXP = 0
                        end
	                    for i=1,lvl-1 do 
                            v:levelUp()
                        end
                        for i=1,v:getNumTrainingSessions() do
                            if i % 3 == 1 then
                                v:incrementStatPower(1)
                            elseif i % 3 == 2 then
                                v:incrementMaxHP(10)
                            else
                                v:incrementStatAwareness(1)
                            end
                        end
                        v.tSaved.nNumTrainingSessions = 0
                        v.tSaved.tStats.nCurMana = v:getMaxMana()
                        v.tSaved.tStats.nCurHP = v:getMaxHP()
                    end

                    g_MiddleManager.tSaved.nStarRatingXP = 0
                    g_MiddleManager.tSaved.nStarRatingLevel = 1
                    g_MiddleManager.tSaved.nLevel = 1

                    while g_MiddleManager.tSaved.nStarRatingLevel < lvl do
		                local nXPForNextLevel = g_MiddleManager:getStarRatingXPForNextLevel()
                        g_MiddleManager:incrementStarRatingXP(nXPForNextLevel)
                    end
                    while g_MiddleManager.tSaved.nLevel < lvl and g_MiddleManager.tSaved.nLevel < MiddleManager.kMAX_LEVEL do
                        g_MiddleManager:incrementLevel(1)
                    end
                    --[[
                    local mm = g_MiddleManager:getAvatar()
                    mm.tSaved.tStats.nLevel = 1
                    --mm.tSaved.nNumTrainingSessions = 0
                    for i=1,lvl-1 do 
                        mm:levelUp()
                    end
                    ]]--
                end
            end
            self.DBG_awaitingInput = nil
        elseif key == 122 then -- Z
            local x, y = Renderer:getCurZoomScl()
            print('zoom '..x..' '..y)
            x = x - nZoomIncrement
            y = y - nZoomIncrement
            Renderer.setCameraScl(x, y, nDuration)
        elseif key == 120 then
            local x, y = Renderer:getCurZoomScl()
            print('zooooom '..x..' '..y)
            x = x + nZoomIncrement
            y = y + nZoomIncrement
            Renderer.setCameraScl(x, y, nDuration)            
            --Renderer.setCameraScl(1, 1, nDuration)
        elseif key == 119 then -- W
            if g_Game:getActiveHeroes()[1] then
                local rTile, nRow, nCol = g_Game:getActiveHeroes()[1]:getOccupyingTile('tiletype')
                if nRow and nCol then
                    nRow = nRow - 1
                    if nRow < 1 then
                        nRow = 1
                    end
                end
                local x, y = g_CurrentMap:getWorldCoords(nRow, nCol)
                if x then                    
                    g_Game:getActiveHeroes()[1]:setPosition(x, y, true)
                end
            end
        elseif key == 115 then -- S
            if g_Game:getActiveHeroes()[1] then
                local rTile, nRow, nCol = g_Game:getActiveHeroes()[1]:getOccupyingTile('tiletype')
                if nRow and nCol then
                    nRow = nRow + 1
                end
                local x, y = g_CurrentMap:getWorldCoords(nRow, nCol)
                if x then
                    g_Game:getActiveHeroes()[1]:setPosition(x, y, true)
                end
            end
        elseif key == 97 then -- A
            if g_Game:getActiveHeroes()[1] then
                local rTile, nRow, nCol = g_Game:getActiveHeroes()[1]:getOccupyingTile('tiletype')
                if nRow and nCol then
                    nCol = nCol - 1
                    if nCol < 1 then
                        nCol = 1
                    end
                end
                local x, y = g_CurrentMap:getWorldCoords(nRow, nCol)
                if x then
                    g_Game:getActiveHeroes()[1]:setPosition(x, y, true)
                end
            end
        elseif key == 100 then -- D
            if g_Game:getActiveHeroes()[1] then
                local rTile, nRow, nCol = g_Game:getActiveHeroes()[1]:getOccupyingTile('tiletype')
                if nRow and nCol then
                    nCol = nCol + 1
                end
                local x, y = g_CurrentMap:getWorldCoords(nRow, nCol)
                if x then
                    g_Game:getActiveHeroes()[1]:setPosition(x, y, true)
                end
            end
        elseif key == 99 then -- C
            local cheatAmount = 300
            g_Game:addCash(cheatAmount)
            DFAnalytics.logEvent("csh_ernd", { type = "chet", amount = cheatAmount })
            DFAnalytics.incrStat('csh_ernd_chet', cheatAmount)
        elseif key == 112 then -- P
            local cheatAmount = 50
            g_Game:addPP(cheatAmount)
            DFAnalytics.logEvent("ppp_ernd", { type = "chet", amount = cheatAmount })
            DFAnalytics.incrStat('ppp_ernd_chet', cheatAmount)
        elseif key == 114 then -- R
            g_MiddleManager:incrementStarRatingXP(100)
        elseif key == 111 then -- O
            g_DebugPatrolTimeSkip = true
        elseif key == 105 then -- I
            g_Game:getActiveHeroes()[1]:incrementStatPower(1000)
            g_Game:getActiveHeroes()[1]:incrementMaxHP(1000)
            g_Game:getActiveHeroes()[1]:setCurHP(1000)
        elseif key == 109 then -- M
            g_MiddleManager:incrementLevel(1)
            g_City:setUnlockLevel(g_MiddleManager:getLevel())            
        elseif key == 108 then -- L
            for i, rHero in ipairs(g_Game:getActiveHeroes()) do
                rHero:incrementXP(rHero:getXPForNextLevel()*.5)
            end
            g_MiddleManager:incrementStarRatingXP(g_MiddleManager:getStarRatingXPForNextLevel() - g_MiddleManager:getStarRatingXP()-1)
        elseif key == 118 then -- V
            g_Game:saveGame()
        elseif key == 98 then -- B
            g_Game:clearSaveGame()
        elseif key == 43 then -- + on numpad
            g_MiddleManager:incrementStarRatingXP(g_MiddleManager:getStarRatingXPForNextLevel())
        elseif key == string.byte("P") then -- Pause
            g_Game:setDebugPaused(not g_Game:getDebugPaused())
        elseif key == string.byte("L") then -- set level
            self.DBG_awaitingInput = key
        elseif key == string.byte("E") then -- Event!
            local ters = g_City:getTerritories()
            local ter = ters[next(ters)]
            local tHeroes = g_Game:getActiveHeroes()
            local ev = ter:addEvent("firerescuelvl1_2")
            ev:addHeroToEvent(tHeroes[1])
            g_eventManager:queueEvent(ev)
            g_eventManager:startNextEventInQueue()
        elseif key == string.byte("O") then -- Lose!
            local ev = g_eventManager:getCurEvent()

            local en = ev:getAllHeroActors()
            for i,v in ipairs(en) do
                if not v:isDead() then
                    v:takeDamage(999999)
                end
            end
        elseif key == string.byte("R") then -- robot battle
            local rTerritory = g_City:getTerritory('StabbersRange')
            rTerritory:setEventToGenerateOverride('boxingMissile_3')
            rTerritory:generateEvents()
            rTerritory:setUnlocked(true)
        elseif key == string.byte("W") then -- Win!
            local ev = g_eventManager:getCurEvent()

            ev.tRewardOverride = {WinXP=999999}
            local en = ev:getAllEnemies()
            for i,v in ipairs(en) do
                if not v:isDead() then
                    v:takeDamage(999999)
                end
            end
        elseif key == string.byte("F") then -- Fight!
            local ters = g_City:getTerritories()
            local ter = ters[next(ters)]

            local tHeroes = g_Game:getActiveHeroes()
            local maxLevel = 1

            for _, rHero in ipairs(tHeroes) do
                if not rHero:isAssignedToEvent() then
                    maxLevel = math.max(maxLevel, rHero:getStatLevel())
                end
            end
            maxLevel = math.min(math.max(1,math.ceil(maxLevel*.5)), 5)

            local ev = ter:addEvent("robberylvl"..maxLevel.."_3")

            local i = 0
            local maxHeroes = ev:getNumMaxHeroes()
            for _, rHero in ipairs(tHeroes) do
                if not rHero:isAssignedToEvent() then
                    ev:addHeroToEvent(rHero)
                    i = i+1
                    if i >= maxHeroes then break end
                end
            end

            g_eventManager:queueEvent(ev)
            g_eventManager:startNextEventInQueue()
        elseif key == string.byte("T") then -- print lua debug stuff
            if g_DebugUI then
                g_DebugUI:showLuaDebugMetrics()
            else
                print("you must first enable the debug ui!")
            end
        end
    end
--    print(tostring(key).." "..tostring(bDown))
end

return input
