local Renderer = require('Renderer')
local util = require('DFUtil')
local UIElement = require('ui/UIElement')
local ListUI = {}

local kSCROLL_FRICTION = 0.9
local kSCROLL_Y_BUFFER = 150

function ListUI:new(rRenderLayer, tList, rParentUI)
	local Ob = MOAITransform.new()

	function Ob:addItem(rItem)
        for i, testItem in ipairs(self.tItems) do
            if testItem == rItem then
                table.remove(self.tItems, i)
                break
            end
        end
    
		table.insert(self.tItems, rItem)
        rItem:setParentUI(self)
        rItem:setParent(self)
        
        if rItem.getYSize then
            self.nItemYSize = rItem:getYSize() -- assumes uniform size
        end
	end

    function Ob:setScrollingEnabled(bEnabled)
        self.bScrollingEnabled = bEnabled
    end

    function Ob:scrollList(dy)
        if not self.bScrollingEnabled or math.abs(dy) < 0.0001 then
            return
        end
        
        local x, y = self:getLoc()
        local nNewY = y + dy
        if nNewY < self.nOrigY then
            nNewY = self.nOrigY
        end
        local nMaxSize = Renderer.getUIHeight() - self.nOrigY
        
        local itemCount = 0
        if self.tItems then
            for i, rItem in ipairs(self.tItems) do
                if not rItem.bHideFromList then
                    itemCount = itemCount + 1
                end
            end
        end
        
        local nListSize = math.abs(self.nOrigY) + ((itemCount + 1) * self.nItemYSize + kSCROLL_Y_BUFFER)
        local nMaxY = math.abs(nListSize - nMaxSize)
        if nNewY > nMaxY then
            nNewY = nMaxY
        end
        print("GOING TO "..itemCount)
        self:setLoc(x, nNewY)
    end

	function Ob:onFinger(eventType, sx, sy, tProps)
        if self.tItems then
            for i, rItem in ipairs(self.tItems) do
                if rItem.onFinger then
                    rItem:onFinger(eventType, sx, sy, tProps, self.bTouchMoved)
                end
            end
        end
		if eventType == MOAITouchSensor.TOUCH_MOVE and self.bIsDragging then
            if not self.lastSy then self.lastSy = sy end            
			local dy = self.lastSy - sy
            
            local nScreenScaleX, nScreenScaleY = Renderer.getDisplayScl()
            dy = dy / nScreenScaleY
            
            local kTOUCHMOVE_SCROLL_THRESHOLD = 5
			if self.bTouchMoved or math.abs(dy) > kTOUCHMOVE_SCROLL_THRESHOLD then
                self.bTouchMoved = true
                self.lastSy = sy
                self.nScrollVelocity = dy
                self:scrollList(dy)
            end
        elseif eventType == MOAITouchSensor.TOUCH_DOWN then
            self.lastSy = sy
            
            self.bIsDragging = true
        elseif eventType == MOAITouchSensor.TOUCH_UP then      
            self.bTouchMoved = false
            self.bIsDragging = false
        end
		return true
	end

    function Ob:setAnchorLoc(x, y)
        self.nOrigX = x
        self.nOrigY = y
        self:setLoc(x, y)
    end

    function Ob:setTopItem(nIndex)
        if self.tItems[nIndex] then
            self:setLoc(self.nOrigX, self.nOrigY + ((nIndex - 1) * self.nItemYSize))
        end
    end

    function Ob:handleButtonPress(rItem)
        if rItem then
            if self.rCurSelectedItem and self.rCurSelectedItem.onSelected then            
                self.rCurSelectedItem:onSelected(false)
            end
            self.rCurSelectedItem = rItem
            if self.rCurSelectedItem.onSelected then
                self.rCurSelectedItem:onSelected(true)
            end
            if self.rParentUI and self.rParentUI.handleItemSelection then
                self.rParentUI:handleItemSelection(self.rCurSelectedItem)
            end
        end
    end

    function Ob:recalcListPos()
        local nYOffset = 0
        if self.tList then
            for i, rItem in ipairs(self.tList) do
                if not rItem.bHideFromList then
                    rItem:setLoc(0, nYOffset)
                    if rItem.getYSize then
                        nYOffset = nYOffset - rItem:getYSize()
                    end
                end
            end
        end
    end

    function Ob:getSelectedItem()
        return self.rCurSelectedItem
    end

    function Ob:getCurSelectItemIndex()
        for i, rItem in ipairs(self.tItems) do
            if rItem == self.rCurSelectedItem then
                return i
            end
        end
        return 0
    end

    function Ob:getItems()
        return self.tItems
    end

    function Ob:setList(tNewList)
        if not tNewList or table.getn(tNewList) == 0 then
            return
        end

        self.rCurSelectedItem = nil

        if self.tItems then
            for i, rItem in ipairs(self.tItems) do
                rItem:setVisible(false)
            end
        end
    
        self.tItems = {}

        -- initialize the elements in the list
        local nYOffset = 0
        for i, rItem in ipairs(tNewList) do
            self:addItem(rItem)
            rItem:setLoc(0, nYOffset)
            if rItem.getYSize then
                nYOffset = nYOffset - rItem:getYSize()
            end
            self.nItemYSize = rItem:getYSize() -- assumes uniform size
        end
        
        if self.bVisible then
			if self.rRenderLayer and self.tItems then
				for i, rItem in ipairs(self.tItems) do
					rItem:setVisible(true)                   
				end
			end
        end        

        -- HACK
        -- for whichever reason the position and parenting doesn't get updated...
        self:moveLoc(0, 0, 1, 0.1,  MOAIEaseType.LINEAR)
        self:moveLoc(0, 0, -1, 0.1,  MOAIEaseType.LINEAR)
    end

    function Ob:update()
        
        if not self.bIsDragging then
            -- do some momentum!
            self:scrollList(self.nScrollVelocity)
        end
        
        self.nScrollVelocity = self.nScrollVelocity * kSCROLL_FRICTION
        
        if self.tItems then
            for i, rItem in ipairs(self.tItems) do
                if rItem.update then
                    rItem:update()
                end
            end
        end
    end

	function Ob:setVisible(bVisible)
        self.bVisible = bVisible
		if bVisible then
			if self.rRenderLayer and self.tItems then
				for i, rItem in ipairs(self.tItems) do
					rItem:setVisible(true)
				end
			end
		else
			if self.rRenderLayer and self.tItems then
				for i, rItem in ipairs(self.tItems) do
					rItem:setVisible(false)
				end
			end
		end
	end

    function Ob:onPaneClosed()
        if self.tItems then
            for i, rItem in ipairs(self.tItems) do
                if rItem.onPaneClosed then
                    rItem:onPaneClosed()
                end
            end
        end  
    end

	Ob.rRenderLayer = rRenderLayer
	Ob.tList = tList
	Ob.nCurMinIndex = 1
	Ob.rSelectedItem = nil
    Ob.nItemYSize = 0
    Ob.rParentUI = rParentUI
    Ob.nOrigX = 0
    Ob.nOrigY = 0
    Ob.nScrollVelocity = 0
    Ob.bIsDragging = false

    Ob.bScrollingEnabled = true

    if tList then
        Ob:setList(tList)
    end

	return Ob
end

return ListUI